"""

A reference internal for Diffusion Entropy Reducer graph clustering algorithm. See

M. Kozdoba and S. Mannor, Community Detection via Measure Space Embedding, NIPS 2015


Code from: https://github.com/komarkdev/der_graph_clustering

The main function is

der_graph_clustering(G, NCOMPONENTS = 2,
                            WALK_LEN = 3,
                            alg_threshold = None,
                            alg_iterbound = 100,
                            do_soften = True
                        )
Arguments:
G - an undirected networkx graph object,
The rest of the parameters are as described in the paper.
Usage example is in block_model_test.py


Code structure:
* Kmeans class implements a generic K-means like skeleton.
* WeightedMeasNodes encapuslates the specific likelihood computations.
* der_graph_clustering is the main function.
  It creates the walks, creates an initialization, runs the algorithm,
  and finally extracts the communities.


"""

import numpy as np
import scipy
import networkx as nx


def _multiply_matrix_rows(mults, M):
    N = M.shape[0]
    diag = scipy.sparse.dia_matrix((mults.reshape((1, N)), np.array([0])), shape=(N, N))
    return diag.dot(M)


class KMeans(object):
    STRICT_INCREASE_FLAG = True

    def __init__(self, init_params, data, node_implementation):
        assert scipy.sparse.isspmatrix_csr(
            data
        ), "data should be scipy sparse csr matrix"

        THR = 0.00000001

        assert (
            max(np.abs(data.sum(axis=1) - 1)) < THR
        ), "Non probabilities on input! - {}".format(max(np.abs(data.sum(axis=1) - 1)))

        data = _multiply_matrix_rows(1 / data.sum(axis=1), data)

        assert max(np.abs(data.sum(axis=1) - 1)) < THR, "Non probabilities on input!"
        assert (
            max(np.abs(init_params.sum(axis=1) - 1)) < THR
        ), "Non probabilities on params!"
        init_params = init_params / (
            init_params.sum(axis=1).reshape(init_params.shape[0], 1)
        )
        assert (
            max(np.abs(init_params.sum(axis=1) - 1)) < THR
        ), "Non probabilities on params!"

        self.params = init_params
        self.ncomps = init_params.shape[0]

        self.node_implementation = node_implementation
        self.data = data
        self.N = data.shape[0]

        self.Q = np.zeros((self.N, self.ncomps))
        self.node_probabilities = np.zeros((self.N, self.ncomps))

    def computeQ(self):

        for i in range(self.ncomps):
            self.node_probabilities[
                :, i
            ] = self.node_implementation.node_log_probabilities(
                self.data, self.params[i]
            )
        max_idxs = np.argmax(self.node_probabilities, axis=1)

        self.Q = np.zeros((self.N, self.ncomps))
        dist_count = 0

        for i in range(self.N):
            self.Q[i, max_idxs[i]] = 1.0
            dist_count += self.node_probabilities[i, max_idxs[i]]

        return dist_count

    def optimize_step(self):
        self.params = self.node_implementation.optimize(self.data, self.Q)

    def optimize(self, threshold, iterbound=100):

        self.loglikelihood = self.computeQ()

        step_cnt = 1
        while True:

            if step_cnt > iterbound:
                break

            self.optimize_step()

            loglikelihood = self.computeQ()

            if not self.STRICT_INCREASE_FLAG:
                likelihood_diff = np.abs(loglikelihood - self.loglikelihood)
            else:
                likelihood_diff = loglikelihood - self.loglikelihood
                assert likelihood_diff > -1.0e-10, "Likelihood decrease!! : {}".format(
                    likelihood_diff
                )

            self.loglikelihood = loglikelihood

            step_cnt += 1

            if (threshold != None) and (likelihood_diff < threshold):
                break

        return


def __rand_measure(k, smoother=0.01):
    rm = np.random.random(size=k) + smoother
    rm /= rm.sum()
    return rm


class WeightedMeasNodes(object):
    def __init__(self, weights, k):

        self.k = k
        self.kzeros = np.zeros(k)
        N = weights.shape[0]
        self.weights = weights.reshape((N, 1))

    def node_log_probabilities__(self, data, param):

        N = data.shape[0]
        k = self.k
        log_param = np.log(param)
        zero_idx = log_param == -np.inf
        log_param[zero_idx] = 0.0
        res = data.dot(log_param.reshape((k, 1))).reshape((N,))
        self.kzeros[zero_idx] = 1
        non_abs_cont = (data.dot(self.kzeros.reshape((k, 1))) > 0).reshape((N,))
        self.kzeros[zero_idx] = 0
        res[non_abs_cont] = -np.inf
        return res

    def node_log_probabilities(self, data, param):

        w = self.weights.reshape((data.shape[0],))
        log_probs = self.node_log_probabilities__(data, param)
        inf_idx = log_probs == -np.inf
        log_probs[inf_idx] = 0
        log_probs = (log_probs * w).reshape((data.shape[0],))
        log_probs[inf_idx] = -np.inf
        return log_probs.reshape((data.shape[0],))

    def optimize__(self, data, Q):
        k = self.k
        ncomp = Q.shape[1]
        params = np.zeros((ncomp, k))
        empty_components = []
        for i in range(ncomp):
            s = Q[:, i].sum()
            if s > 0:
                pos_idx = Q[:, i] > 0
                params[i, :] = _multiply_matrix_rows(
                    Q[pos_idx, i] / s, data[pos_idx, :]
                ).sum(axis=0)
            else:
                empty_components.append(i)

        assert len(empty_components) != ncomp, "All components empty!"

        for i in empty_components:
            params[i, :] = 0.0

        return params

    def optimize(self, data, Q):
        # this currently assumes data is very specific, i.e. of the length of weights
        return self.optimize__(data, self.weights * Q)

    def get_communities(self, params, data):

        communities = []
        NCOMPONENTS = params.shape[0]
        node_log_probs = np.zeros((self.k, NCOMPONENTS))
        for i in range(NCOMPONENTS):
            node_log_probs[:, i] = self.node_log_probabilities(data, params[i])

        labels = np.argmax(node_log_probs, axis=1)
        for i in range(NCOMPONENTS):
            communities.append(list(np.arange(self.k)[labels == i]))

        return communities

    def init_params_soften(self, params, alpha=0.000001):

        ncomp, k = params.shape
        unif = np.ones(k) / k

        return (1 - alpha) * params + alpha * unif.reshape((1, k))

    def init_params_random_subset_data(self, ncomp, data, weights=None):

        Ndata = data.shape[0]
        params = np.zeros((ncomp, Ndata))

        if weights is None:
            weights = np.ones(Ndata)

        step = Ndata / ncomp

        for i in range(ncomp):

            if i == ncomp - 1:
                params[i, int(i * step) :] = 1.0
            else:
                params[i, int(i * step) : int((i + 1) * step)] = 1.0

        perm_idx = np.random.permutation(Ndata)
        params = params[:, perm_idx]

        params = params * weights.reshape((1, Ndata))
        params = params / params.sum(axis=1).reshape((ncomp, 1))

        params = ((data.T).dot(params.T)).T

        return params


def __graph_transition_matrix(G, sparse=True):
    A = nx.adjacency_matrix(G).astype("float")
    # normalize rows to sum to 1
    degs = A.sum(axis=1)

    # take care of zero degree
    degs[degs == 0] = 1

    N = len(degs)

    if sparse == True:
        rev_degs = 1 / degs
        diag = scipy.sparse.dia_matrix(
            (rev_degs.reshape((1, N)), np.array([0])), shape=(N, N)
        )
        A = diag.dot(A)
    else:
        A = A.todense()
        A = A / degs.reshape((A.shape[0], 1))

    return A


def __create_walks(TM, WALK_LEN):
    # Should be faster, TM is sparse

    N = TM.shape[0]

    powers = [TM]
    for i in range(1, WALK_LEN):
        powers.append(powers[-1].dot(TM))

    totals = scipy.sparse.csr_matrix((N, N))
    for m in powers:
        totals = totals + m

    totals = totals / WALK_LEN

    return totals


def der_graph_clustering(
    graph,
    ncomponents=2,
    walk_len=3,
    alg_threshold=None,
    alg_iterbound=100,
    do_soften=True,
):

    TM = __graph_transition_matrix(graph, sparse=True)
    graph_size = TM.shape[0]

    degs = graph.degree()
    weights = np.array(list(map(lambda i: degs[i], graph.nodes())))

    assert sum(weights > 0) == len(weights), "Zero weights found!"

    data = __create_walks(TM, walk_len)

    MN = WeightedMeasNodes(weights, k=graph_size)

    init_params = MN.init_params_random_subset_data(ncomponents, data, weights)

    if do_soften:
        init_params = MN.init_params_soften(init_params, alpha=0.000001)

    alg = KMeans(init_params, data, MN)

    alg.optimize(alg_threshold, iterbound=alg_iterbound)
    communities = MN.get_communities(alg.params, data)

    return communities, alg



import sys
import numpy as np
from typing import Callable
from copy import deepcopy
from cdlib.algorithms.internal import DER

# import community as louvain_modularity
from community import community_louvain
from collections import defaultdict
from cdlib import NodeClustering, FuzzyNodeClustering
from cdlib.algorithms.internal.belief_prop import detect_belief_communities
from cdlib.algorithms.internal.em import EM_nx
from cdlib.algorithms.internal.scan import SCAN_nx
from cdlib.algorithms.internal.GDMP2_nx import GDMP2
from cdlib.algorithms.internal.AGDL import Agdl
from cdlib.algorithms.internal.FuzzyCom import fuzzy_comm
from cdlib.algorithms.internal.Markov import markov
from cdlib.algorithms.internal.ga import ga_community_detection
from cdlib.algorithms.internal.SiblinarityAntichain import (
    matrix_node_recursive_antichain_partition,
)
from cdlib.algorithms.internal.LSWL import (
    LSWLCommunityDiscovery_offline,
    LSWLPlusCommunityDetection,
    LSWLCommunityDiscovery,
)
from cdlib.algorithms.internal.spectralCD import spectral_communities
from cdlib.algorithms.internal.modularity_m import ModularityMCommunityDiscovery
from cdlib.algorithms.internal.modularity_r import ModularityRCommunityDiscovery
from cdlib.algorithms.internal.headtail import HeadTail
from cdlib.algorithms.internal.Kcut import kcut_exec
from cdlib.algorithms.internal.paris import paris as paris_alg, paris_best_clustering
from cdlib.algorithms.internal.principled import principled
from cdlib.algorithms.internal.MCODE import m_code
from cdlib.algorithms.internal.RSC import rsc_evaluate_graph
import warnings

import markov_clustering as mc
from chinese_whispers import chinese_whispers as cw
from chinese_whispers import aggregate_clusters
from thresholdclustering.thresholdclustering import best_partition as th_best_partition
import networkx as nx

from cdlib.utils import (
    convert_graph_formats,
    __from_nx_to_graph_tool,
    affiliations2nodesets,
    nx_node_integer_mapping,
)

missing_packages = set()

try:
    import infomap as imp
except ModuleNotFoundError:
    missing_packages.add("infomap")
    imp = None

try:
    from wurlitzer import pipes
except ModuleNotFoundError:
    missing_packages.add("wurlitzer")
    pipes = None

try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None

try:
    import leidenalg
except ModuleNotFoundError:
    missing_packages.add("leidenalg")
    leidenalg = None

try:
    import graph_tool.all as gt
except ModuleNotFoundError:
    missing_packages.add("graph_tool")
    gt = None

try:
    import karateclub
except ModuleNotFoundError:
    missing_packages.add("karateclub")


if len(missing_packages) > 0:
    print(
        "Note: to be able to use all crisp methods, you need to install some additional packages: ",
        missing_packages,
    )

try:
    from GraphRicciCurvature.OllivierRicci import OllivierRicci
except ModuleNotFoundError:
    OllivierRicci = None

try:
    import pycombo as pycombo_part
except ModuleNotFoundError:
    pycombo_part = None

__all__ = [
    "louvain",
    "leiden",
    "rb_pots",
    "rber_pots",
    "cpm",
    "significance_communities",
    "surprise_communities",
    "greedy_modularity",
    "der",
    "label_propagation",
    "async_fluid",
    "infomap",
    "walktrap",
    "girvan_newman",
    "em",
    "scan",
    "gdmp2",
    "spinglass",
    "eigenvector",
    "agdl",
    "frc_fgsn",
    "sbm_dl",
    "sbm_dl_nested",
    "markov_clustering",
    "edmot",
    "chinesewhispers",
    "siblinarity_antichain",
    "ga",
    "belief",
    "threshold_clustering",
    "lswl_plus",
    "lswl",
    "mod_m",
    "mod_r",
    "head_tail",
    "kcut",
    "gemsec",
    "scd",
    "pycombo",
    "paris",
    "principled_clustering",
    "ricci_community",
    "spectral",
    "mcode",
    "r_spectral_clustering",
]


def girvan_newman(g_original: object, level: int) -> NodeClustering:
    """
    The Girvan–Newman algorithm detects communities by progressively removing edges from the original graph.
    The algorithm removes the "most valuable" edge, traditionally the edge with the highest betweenness centrality, at each step. As the graph breaks down into pieces, the tightly knit community structure is exposed and the result can be depicted as a dendrogram.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param level: the level where to cut the dendrogram
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.girvan_newman(G, level=3)

    :References:

    Girvan, Michelle, and Mark EJ Newman. `Community structure in social and biological networks. <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC122977/>`_ Proceedings of the national academy of sciences 99.12 (2002): 7821-7826.
    """

    g = convert_graph_formats(g_original, nx.Graph)

    gn_hierarchy = nx.algorithms.community.girvan_newman(g)
    coms = []
    for _ in range(level):
        coms = next(gn_hierarchy)

    communities = []

    for c in coms:
        communities.append(list(c))

    return NodeClustering(
        communities, g_original, "Girvan Newman", method_parameters={"level": level}
    )


def em(g_original: object, k: int) -> NodeClustering:
    """
    EM is based on based on a mixture model.
    The algorithm uses the expectation–maximization algorithm to detect structure in networks.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        Yes      No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param k: the number of desired communities
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.em(G, k=3)

    :References:

    Newman, Mark EJ, and Elizabeth A. Leicht. `Mixture community and exploratory analysis in networks.  <https://www.pnas.org/content/104/23/9564/>`_  Proceedings of the National Academy of Sciences 104.23 (2007): 9564-9569.

    .. note:: Reference implementation: https://github.com/duckneo/CommunityDetection
    """

    g = convert_graph_formats(g_original, nx.Graph)
    g, maps = nx_node_integer_mapping(g)

    algorithm = EM_nx(g, k)
    coms = algorithm.execute()

    if maps is not None:
        communities = []
        for c in coms:
            communities.append([maps[n] for n in c])
        nx.relabel_nodes(g, maps, False)
    else:
        communities = [list(c) for c in coms]

    return NodeClustering(communities, g_original, "EM", method_parameters={"k": k})


def scan(g_original: object, epsilon: float, mu: int) -> NodeClustering:
    """
    SCAN (Structural Clustering Algorithm for Networks) is an algorithm which detects clusters, hubs and outliers in networks.
    It clusters vertices based on a structural similarity measure.
    The method uses the neighborhood of the vertices as clustering criteria instead of only their direct connections.
    Vertices are grouped into the clusters by how they share neighbors.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param epsilon: the minimum threshold to assigning cluster membership
    :param mu: minimum number of neineighbors with a structural similarity that exceeds the threshold epsilon
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.scan(G, epsilon=0.7, mu=3)

    :References:

    Xu, X., Yuruk, N., Feng, Z., & Schweiger, T. A. (2007, August). `Scan: a structural clustering algorithm for networks. <http://www1.se.cuhk.edu.hk/~hcheng/seg5010/slides/p824-xu.pdf/>`_ In Proceedings of the 13th ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 824-833)
    """

    g = convert_graph_formats(g_original, nx.Graph)

    algorithm = SCAN_nx(g, epsilon, mu)
    coms = algorithm.execute()
    return NodeClustering(
        coms, g_original, "SCAN", method_parameters={"epsilon": epsilon, "mu": mu}
    )


def gdmp2(g_original: object, min_threshold: float = 0.75) -> NodeClustering:
    """
    Gdmp2 is a method for identifying a set of dense subgraphs of a given sparse graph.
    It is inspired by an effective technique designed for a similar problem—matrix blocking, from a different discipline (solving linear systems).


    **Supported Graph Types**

    ========== ======== ======== =========
    Undirected Directed Weighted Bipartite
    ========== ======== ======== =========
    Yes        Yes      No       Yes
    ========== ======== ======== =========

    :param g_original: a networkx/igraph object
    :param min_threshold:  the minimum density threshold parameter to control the density of the output subgraphs, default 0.75
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.gdmp2(G)

    :References:

    Chen, Jie, and Yousef Saad. `Dense subgraph extraction with application to community detection. <https://ieeexplore.ieee.org/document/5677532/>`_ IEEE Transactions on Knowledge and Data Engineering 24.7 (2012): 1216-1230.

    .. note:: Reference implementation: https://github.com/imabhishekl/CSC591_Community_Detection
    """

    g = convert_graph_formats(g_original, nx.Graph)
    g, maps = nx_node_integer_mapping(g)

    coms = GDMP2(g, min_threshold)

    if maps is not None:
        communities = []
        for c in coms:
            communities.append([maps[n] for n in c])
        nx.relabel_nodes(g, maps, False)
    else:
        communities = coms

    return NodeClustering(
        communities,
        g_original,
        "GDMP2",
        method_parameters={"min_threshold": min_threshold},
    )


def spinglass(g_original: object) -> NodeClustering:
    """
    Spinglass relies on an analogy between a very popular statistical mechanic model called Potts spin glass, and the community structure.
    It applies the simulated annealing optimization technique on this model to optimize the modularity.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.spinglass(G)

    :References:

    Reichardt, Jörg, and Stefan Bornholdt. `Statistical mechanics of community detection. <https://journals.aps.org/pre/abstract/10.1103/PhysRevE.74.016110/>`_ Physical Review E 74.1 (2006): 016110.
    """
    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)
    coms = g.community_spinglass()
    communities = []

    for c in coms:
        communities.append([g.vs[x]["name"] for x in c])

    return NodeClustering(
        communities, g_original, "Spinglass", method_parameters={"": ""}
    )


def eigenvector(g_original: object) -> NodeClustering:
    """
    Newman's leading eigenvector method for detecting community structure based on modularity.
    This is the proper internal of the recursive, divisive algorithm: each split is done by maximizing the modularity regarding the original network.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.eigenvector(G)

    :References:

    Newman, Mark EJ. `Finding community structure in networks using the eigenvectors of matrices. <https://journals.aps.org/pre/pdf/10.1103/PhysRevE.74.036104/>`_ Physical review E 74.3 (2006): 036104.
    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)
    coms = g.community_leading_eigenvector()

    communities = [g.vs[x]["name"] for x in coms]

    return NodeClustering(
        communities, g_original, "Eigenvector", method_parameters={"": ""}
    )


def agdl(g_original: object, number_communities: int, kc: int) -> NodeClustering:
    """
    AGDL is a graph-based agglomerative algorithm, for clustering high-dimensional data.
    The algorithm uses  the indegree and outdegree to characterize the affinity between two clusters.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        Yes      Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param number_communities: number of communities
    :param kc: size of the neighbor set for each cluster
    :return: NodeClustering object

     :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.agdl(g, number_communities=3, kc=4)

    :References:

    Zhang, W., Wang, X., Zhao, D., & Tang, X. (2012, October). `Graph degree linkage: Agglomerative clustering on a directed graph. <https://link.springer.com/chapter/10.1007/978-3-642-33718-5_31/>`_ In European Conference on Computer Vision (pp. 428-441). Springer, Berlin, Heidelberg.

    .. note:: Reference implementation: https://github.com/myungjoon/GDL
    """

    g = convert_graph_formats(g_original, nx.Graph)

    communities = Agdl(g, number_communities, kc)
    nodes = {k: v for k, v in enumerate(g.nodes())}
    coms = []
    for com in communities:
        coms.append([nodes[n] for n in com])

    return NodeClustering(
        coms,
        g_original,
        "AGDL",
        method_parameters={"number_communities": number_communities, "kc": kc},
    )


def louvain(
    g_original: object,
    weight: str = "weight",
    resolution: float = 1.0,
    randomize: int = None,
) -> NodeClustering:
    """
    Louvain  maximizes a modularity score for each community.
    The algorithm optimises the modularity in two elementary phases:
    (1) local moving of nodes;
    (2) aggregation of the network.
    In the local moving phase, individual nodes are moved to the community that yields the largest increase in the quality function.
    In the aggregation phase, an aggregate network is created based on the partition obtained in the local moving phase.
    Each community in this partition becomes a node in the aggregate network. The two phases are repeated until the quality function cannot be increased further.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param weight: str, optional the key in graph to use as weight. Default to 'weight'
    :param resolution: double, optional  Will change the size of the communities, default to 1.
    :param randomize: int, RandomState instance or None, optional (default=None). If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.louvain(G, weight='weight', resolution=1.)

    :References:

    Blondel, Vincent D., et al. `Fast unfolding of communities in large networks. <https://iopscience.iop.org/article/10.1088/1742-5468/2008/10/P10008/meta/>`_ Journal of statistical mechanics: theory and experiment 2008.10 (2008): P10008.

    .. note:: Reference implementation: https://github.com/taynaud/python-louvain
    """

    g = convert_graph_formats(g_original, nx.Graph)

    coms = community_louvain.best_partition(
        g, weight=weight, resolution=resolution, randomize=randomize
    )

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in coms.items():
        coms_to_node[c].append(n)

    coms_louvain = [list(c) for c in coms_to_node.values()]
    return NodeClustering(
        coms_louvain,
        g_original,
        "Louvain",
        method_parameters={
            "weight": weight,
            "resolution": resolution,
            "randomize": randomize,
        },
    )


def leiden(
    g_original: object, initial_membership: list = None, weights: list = None
) -> NodeClustering:
    """
    The Leiden algorithm is an improvement of the Louvain algorithm.
    The Leiden algorithm consists of three phases:
    (1) local moving of nodes,
    (2) refinement of the partition
    (3) aggregation of the network based on the refined partition, using the non-refined partition to create an initial partition for the aggregate network.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param initial_membership:  list of int Initial membership for the partition. If :obj:`None` then defaults to a singleton partition. Deafault None
    :param weights: list of double, or edge attribute Weights of edges. Can be either an iterable or an edge attribute. Deafault None
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.leiden(G)

    :References:

    Traag, Vincent, Ludo Waltman, and Nees Jan van Eck. `From Louvain to Leiden: guaranteeing well-connected communities. <https://arxiv.org/abs/1810.08473/>`_ arXiv preprint arXiv:1810.08473 (2018).

    .. note:: Reference implementation: https://github.com/vtraag/leidenalg
    """
    global leidenalg
    if ig is None or leidenalg is None:
        try:
            import leidenalg
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install igraph and leidenalg to use the "
                "selected feature."
            )

    g = convert_graph_formats(g_original, ig.Graph)

    part = leidenalg.find_partition(
        g,
        leidenalg.ModularityVertexPartition,
        initial_membership=initial_membership,
        weights=weights,
    )
    coms = [g.vs[x]["name"] for x in part]
    return NodeClustering(
        coms,
        g_original,
        "Leiden",
        method_parameters={
            "initial_membership": initial_membership,
            "weights": weights,
        },
    )


def rb_pots(
    g_original: object,
    initial_membership: list = None,
    weights: list = None,
    resolution_parameter: float = 1,
) -> NodeClustering:
    """
    Rb_pots is a model where the quality function to optimize is:

    .. math:: Q = \\sum_{ij} \\left(A_{ij} - \\gamma \\frac{k_i k_j}{2m} \\right)\\delta(\\sigma_i, \\sigma_j)

    where :math:`A` is the adjacency matrix, :math:`k_i` is the (weighted) degree of node :math:`i`, :math:`m` is the total number of edges (or total edge weight), :math:`\\sigma_i` denotes the community of node :math:`i` and :math:`\\delta(\\sigma_i, \\sigma_j) = 1` if :math:`\\sigma_i = \\sigma_j` and `0` otherwise.
    For directed graphs a slightly different formulation is used, as proposed by Leicht and Newman :

    .. math:: Q = \\sum_{ij} \\left(A_{ij} - \\gamma \\frac{k_i^\mathrm{out} k_j^\mathrm{in}}{m} \\right)\\delta(\\sigma_i, \\sigma_j),

    where :math:`k_i^\\mathrm{out}` and :math:`k_i^\\mathrm{in}` refers to respectively the outdegree and indegree of node :math:`i` , and :math:`A_{ij}` refers to an edge from :math:`i` to :math:`j`.
    Note that this is the same of Leiden algorithm when setting :math:`\\gamma=1` and normalising by :math:`2m`, or :math:`m` for directed graphs.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        Yes      Yes
    ========== ======== ========


    :param g_original: a networkx/igraph object
    :param initial_membership:  list of int Initial membership for the partition. If :obj:`None` then defaults to a singleton partition. Deafault None
    :param weights: list of double, or edge attribute Weights of edges. Can be either an iterable or an edge attribute. Deafault None
    :param resolution_parameter: double >0 A parameter value controlling the coarseness of the clustering. Higher resolutions lead to more communities, while lower resolutions lead to fewer communities. Default 1
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.rb_pots(G)

    :References:

    Reichardt, J., & Bornholdt, S. (2006).  `Statistical mechanics of community detection. <https://journals.aps.org/pre/abstract/10.1103/PhysRevE.74.016110/>`_  Physical Review E, 74(1), 016110. 10.1103/PhysRevE.74.016110

    Leicht, E. A., & Newman, M. E. J. (2008).  `Community Structure in Directed Networks. <https://www.ncbi.nlm.nih.gov/pubmed/18517839/>`_  Physical Review Letters, 100(11), 118703. 10.1103/PhysRevLett.100.118703

    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)

    part = leidenalg.find_partition(
        g,
        leidenalg.RBConfigurationVertexPartition,
        resolution_parameter=resolution_parameter,
        initial_membership=initial_membership,
        weights=weights,
    )
    coms = [g.vs[x]["name"] for x in part]
    return NodeClustering(
        coms,
        g_original,
        "RB Pots",
        method_parameters={
            "initial_membership": initial_membership,
            "weights": weights,
            "resolution_parameter": resolution_parameter,
        },
    )


def rber_pots(
    g_original: object,
    initial_membership: list = None,
    weights: list = None,
    node_sizes: list = None,
    resolution_parameter: float = 1,
) -> NodeClustering:
    """
    rber_pots is a  model where the quality function to optimize is:

    .. math:: Q = \\sum_{ij} \\left(A_{ij} - \\gamma p \\right)\\delta(\\sigma_i, \\sigma_j)

    where :math:`A` is the adjacency matrix,  :math:`p = \\frac{m}{\\binom{n}{2}}` is the overall density of the graph, :math:`\\sigma_i` denotes the community of node :math:`i`, :math:`\\delta(\\sigma_i, \\sigma_j) = 1` if  :math:`\\sigma_i = \\sigma_j` and `0` otherwise, and, finally :math:`\\gamma` is a resolution parameter.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========


    :param g_original: a networkx/igraph object
    :param initial_membership:  list of int Initial membership for the partition. If :obj:`None` then defaults to a singleton partition. Deafault None
    :param weights: list of double, or edge attribute Weights of edges. Can be either an iterable or an edge attribute. Deafault None
    :param node_sizes: list of int, or vertex attribute Sizes of nodes are necessary to know the size of communities in aggregate graphs. Usually this is set to 1 for all nodes, but in specific cases  this could be changed. Deafault None
    :param resolution_parameter: double >0 A parameter value controlling the coarseness of the clustering. Higher resolutions lead to more communities, while lower resolutions lead to fewer communities. Deafault 1
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.rber_pots(G)

    :References:

    Reichardt, J., & Bornholdt, S. (2006).  `Statistical mechanics of community detection. <https://journals.aps.org/pre/abstract/10.1103/PhysRevE.74.016110/>`_  Physical Review E, 74(1), 016110. 10.1103/PhysRevE.74.016110


    .. note:: Reference implementation: https://github.com/vtraag/leidenalg

    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)

    part = leidenalg.find_partition(
        g,
        leidenalg.RBERVertexPartition,
        resolution_parameter=resolution_parameter,
        initial_membership=initial_membership,
        weights=weights,
        node_sizes=node_sizes,
    )
    coms = [g.vs[x]["name"] for x in part]
    return NodeClustering(
        coms,
        g_original,
        "RBER Pots",
        method_parameters={
            "initial_membership": initial_membership,
            "weights": weights,
            "node_sizes": node_sizes,
            "resolution_parameter": resolution_parameter,
        },
    )


def cpm(
    g_original: object,
    initial_membership: list = None,
    weights: list = None,
    node_sizes: list = None,
    resolution_parameter: float = 1,
) -> NodeClustering:
    """
    CPM is a model where the quality function to optimize is:

    .. math:: Q = \\sum_{ij} \\left(A_{ij} - \\gamma \\right)\\delta(\\sigma_i, \\sigma_j)

    where :math:`A` is the adjacency matrix, :math:`\\sigma_i` denotes the community of node :math:`i`, :math:`\\delta(\\sigma_i, \\sigma_j) = 1` if  :math:`\\sigma_i = \\sigma_j` and `0` otherwise, and, finally :math:`\\gamma` is a resolution parameter.

    The internal density of communities

    .. math:: p_c = \\frac{m_c}{\\binom{n_c}{2}} \\geq \\gamma

    is higher than :math:`\\gamma`, while the external density

    :math:`p_{cd} = \\frac{m_{cd}}{n_c n_d} \\leq \\gamma`    is lower than :math:`\\gamma`. In other words, choosing a particular
    :math:`\\gamma` corresponds to choosing to find communities of a particular
    density, and as such defines communities. Finally, the definition of a community is in a sense independent of the actual graph, which is not the case for any of the other methods.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========


    :param g_original: a networkx/igraph object
    :param initial_membership:  list of int Initial membership for the partition. If :obj:`None` then defaults to a singleton partition. Deafault None
    :param weights: list of double, or edge attribute Weights of edges. Can be either an iterable or an edge attribute. Deafault None
    :param node_sizes: list of int, or vertex attribute Sizes of nodes are necessary to know the size of communities in aggregate graphs. Usually this is set to 1 for all nodes, but in specific cases  this could be changed. Deafault None
    :param resolution_parameter: double >0 A parameter value controlling the coarseness of the clustering. Higher resolutions lead to more communities, while lower resolutions lead to fewer communities. Deafault 1
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.cpm(G)

    :References:

    Traag, V. A., Van Dooren, P., & Nesterov, Y. (2011).  `Narrow scope for resolution-limit-free community detection. <https://journals.aps.org/pre/abstract/10.1103/PhysRevE.84.016114/>`_ Physical Review E, 84(1), 016114. 10.1103/PhysRevE.84.016114


    .. note:: Reference implementation: https://github.com/vtraag/leidenalg

    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)

    part = leidenalg.find_partition(
        g,
        leidenalg.CPMVertexPartition,
        resolution_parameter=resolution_parameter,
        initial_membership=initial_membership,
        weights=weights,
        node_sizes=node_sizes,
    )
    coms = [g.vs[x]["name"] for x in part]
    return NodeClustering(
        coms,
        g_original,
        "CPM",
        method_parameters={
            "initial_membership": initial_membership,
            "weights": weights,
            "node_sizes": node_sizes,
            "resolution_parameter": resolution_parameter,
        },
    )


def significance_communities(
    g_original: object, initial_membership: list = None, node_sizes: list = None
) -> NodeClustering:
    """
    Significance_communities is a model where the quality function to optimize is:

    .. math:: Q = \\sum_c \\binom{n_c}{2} D(p_c \\parallel p)

    where :math:`n_c` is the number of nodes in community :math:`c`, :math:`p_c = \\frac{m_c}{\\binom{n_c}{2}}`, is the density of community :math:`c`,  :math:`p = \\frac{m}{\\binom{n}{2}}`  is the overall density of the graph, and finally  :math:`D(x \\parallel y) = x \\ln \\frac{x}{y} + (1 - x) \\ln \\frac{1 - x}{1 - y}` is the binary Kullback-Leibler divergence.
    For directed graphs simply multiply the binomials by 2. The expected Significance in Erdos-Renyi graphs behaves roughly as :math:`\\frac{1}{2} n \\ln n` for both directed and undirected graphs in this formulation.

    .. warning:: This method is not suitable for weighted graphs.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========


    :param g_original: a networkx/igraph object
    :param initial_membership:  list of int Initial membership for the partition. If :obj:`None` then defaults to a singleton partition. Deafault None
    :param node_sizes: list of int, or vertex attribute Sizes of nodes are necessary to know the size of communities in aggregate graphs. Usually this is set to 1 for all nodes, but in specific cases  this could be changed. Deafault None
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.significance_communities(G)

    :References:

    Traag, V. A., Krings, G., & Van Dooren, P. (2013). `Significant scales in community structure. <https://www.nature.com/articles/srep02930/>`_  Scientific Reports, 3, 2930. `10.1038/srep02930 <http://doi.org/10.1038/srep02930>`

    .. note:: Reference implementation: https://github.com/vtraag/leidenalg

    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)

    part = leidenalg.find_partition(
        g,
        leidenalg.SignificanceVertexPartition,
        initial_membership=initial_membership,
        node_sizes=node_sizes,
    )
    coms = [g.vs[x]["name"] for x in part]
    return NodeClustering(
        coms,
        g_original,
        "Significance",
        method_parameters={
            "initial_membership": initial_membership,
            "node_sizes": node_sizes,
        },
    )


def surprise_communities(
    g_original: object,
    initial_membership: list = None,
    weights: list = None,
    node_sizes: list = None,
) -> NodeClustering:
    """

    Surprise_communities is a model where the quality function to optimize is:

    .. math:: Q = m D(q \\parallel \\langle q \\rangle)

    where :math:`m` is the number of edges,  :math:`q = \\frac{\\sum_c m_c}{m}`,  is the fraction of internal edges, :math:`\\langle q \\rangle = \\frac{\\sum_c \\binom{n_c}{2}}{\\binom{n}{2}}` is the expected fraction of internal edges, and finally

    :math:`D(x \\parallel y) = x \\ln \\frac{x}{y} + (1 - x) \\ln \\frac{1 - x}{1 - y}`  is the binary Kullback-Leibler divergence.

    For directed graphs we can multiplying the binomials by 2, and this leaves :math:`\\langle q \\rangle` unchanged, so that we can simply use the same
    formulation.  For weighted graphs we can simply count the total internal weight instead of the total number of edges for :math:`q` , while :math:`\\langle q \\rangle` remains unchanged.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        Yes      Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param initial_membership:  list of int Initial membership for the partition. If :obj:`None` then defaults to a singleton partition. Deafault None
    :param weights: list of double, or edge attribute Weights of edges. Can be either an iterable or an edge attribute. Deafault None
    :param node_sizes: list of int, or vertex attribute Sizes of nodes are necessary to know the size of communities in aggregate graphs. Usually this is set to 1 for all nodes, but in specific cases  this could be changed. Deafault None
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.surprise_communities(G)

    :References:

    Traag, V. A., Aldecoa, R., & Delvenne, J.-C. (2015).  `Detecting communities using asymptotical surprise. <https://journals.aps.org/pre/abstract/10.1103/PhysRevE.92.022816/>`_ Physical Review E, 92(2), 022816. 10.1103/PhysRevE.92.022816

    .. note:: Reference implementation: https://github.com/vtraag/leidenalg

    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)

    part = leidenalg.find_partition(
        g,
        leidenalg.SurpriseVertexPartition,
        initial_membership=initial_membership,
        weights=weights,
        node_sizes=node_sizes,
    )
    coms = [g.vs[x]["name"] for x in part]
    return NodeClustering(
        coms,
        g_original,
        "Surprise",
        method_parameters={
            "initial_membership": initial_membership,
            "weights": weights,
            "node_sizes": node_sizes,
        },
    )


def greedy_modularity(g_original: object, weight: list = None) -> NodeClustering:
    """
    The CNM algorithm uses the modularity to find the communities strcutures.
    At every step of the algorithm two communities that contribute maximum positive value to global modularity are merged.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param weight: list of double, or edge attribute Weights of edges. Can be either an iterable or an edge attribute. Deafault None
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.greedy_modularity(G)

    :References:

    Clauset, A., Newman, M. E., & Moore, C. `Finding community structure in very large networks. <http://ece-research.unm.edu/ifis/papers/community-moore.pdf/>`_ Physical Review E 70(6), 2004
    """
    g = convert_graph_formats(g_original, nx.Graph)

    coms = nx.algorithms.community.greedy_modularity_communities(g, weight)
    coms = [list(x) for x in coms]
    return NodeClustering(
        coms, g_original, "Greedy Modularity", method_parameters={"weight": weight}
    )


def infomap(g_original: object, flags: str = "") -> NodeClustering:
    """
    Infomap is based on ideas of information theory.
    The algorithm uses the probability flow of random walks on a network as a proxy for information flows in the real system and it decomposes the network into modules by compressing a description of the probability flow.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        Yes      Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param flags: str flags for Infomap
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.infomap(G)

    :References:

    Rosvall M, Bergstrom CT (2008) `Maps of random walks on complex networks reveal community structure. <https://www.pnas.org/content/105/4/1118/>`_ Proc Natl Acad SciUSA 105(4):1118–1123

    .. note:: Reference implementation: https://pypi.org/project/infomap/

    .. note:: Infomap Python API documentation: https://mapequation.github.io/infomap/python/
    """
    global imp, pipes
    if imp is None:
        try:
            import infomap as imp
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install infomap to use the selected feature."
            )
    if pipes is None:
        try:
            from wurlitzer import pipes
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install package wurlitzer to use infomap."
            )

    g = convert_graph_formats(g_original, nx.Graph, directed=g_original.is_directed())

    if g_original.is_directed() and "-d" not in flags and "--directed" not in flags:
        flags += " -d"

    g1 = nx.convert_node_labels_to_integers(g, label_attribute="name")
    name_map = nx.get_node_attributes(g1, "name")
    coms_to_node = defaultdict(list)

    with pipes():
        im = imp.Infomap(flags)

        if tuple(map(int, imp.__version__.split("."))) >= (1, 7, 1):
            n_dict = {i: str(n) for i, n in enumerate(g1.nodes)}
            im.add_nodes(n_dict)
        else:
            im.add_nodes(g1.nodes)

        for source, target, data in g1.edges(data=True):
            if "weight" in data:
                im.add_link(source, target, data["weight"])
            else:
                im.add_link(source, target)
        im.run()

        for node_id, module_id in im.modules:
            node_name = name_map[node_id]
            coms_to_node[module_id].append(node_name)

    coms_infomap = [list(c) for c in coms_to_node.values()]
    return NodeClustering(
        coms_infomap, g_original, "Infomap", method_parameters={"flags": flags}
    )


def walktrap(g_original: object) -> NodeClustering:
    """
    walktrap is an approach based on random walks.
    The general idea is that if you perform random walks on the graph, then the walks are more likely to stay within the same community because there are only a few edges that lead outside a given community. Walktrap runs short random walks and uses the results of these random walks to merge separate communities in a bottom-up manner.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClusterint object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.walktrap(G)

    :References:

    Pons, Pascal, and Matthieu Latapy. `Computing communities in large networks using random walks. <http://jgaa.info/accepted/2006/PonsLatapy2006.10.2.pdf/>`_ J. Graph Algorithms Appl. 10.2 (2006): 191-218.
    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)
    coms = g.community_walktrap().as_clustering()
    communities = []

    for c in coms:
        communities.append([g.vs[x]["name"] for x in c])

    return NodeClustering(
        communities, g_original, "Walktrap", method_parameters={"": ""}
    )


def label_propagation(g_original: object) -> NodeClustering:
    """
    The Label Propagation algorithm (LPA) detects communities using network structure alone.
    The algorithm doesn’t require a pre-defined objective function or prior information about the communities.
    It works as follows:
    -Every node is initialized with a unique label (an identifier)
    -These labels propagate through the network
    -At every iteration of propagation, each node updates its label to the one that the maximum numbers of its neighbours belongs to. Ties are broken uniformly and randomly.
    -LPA reaches convergence when each node has the majority label of its neighbours.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: EdgeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.label_propagation(G)

    :References:

    Cordasco, G., & Gargano, L. (2010, December). Community detection via semi-synchronous label propagation algorithms. In 2010 IEEE international workshop on: business applications of social network analysis (BASNA) (pp. 1-8). IEEE.
    """

    g = convert_graph_formats(g_original, nx.Graph)

    coms = list(nx.algorithms.community.label_propagation_communities(g))
    coms = [list(x) for x in coms]

    return NodeClustering(
        coms, g_original, "Label Propagation", method_parameters={"": ""}
    )


def async_fluid(g_original: object, k: int) -> NodeClustering:
    """
    Fluid Communities (FluidC) is based on the simple idea of fluids (i.e., communities) interacting in an environment (i.e., a non-complete graph), expanding and contracting.
    It is propagation-based algorithm and it allows to specify the number of desired communities (k) and it is asynchronous, where each vertex update is computed using the latest partial state of the graph.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========


    :param g_original: a networkx/igraph object
    :param k: Number of communities to search
    :return: EdgeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.async_fluid(G,k=2)


    :References:

    Ferran Parés, Dario Garcia-Gasulla, Armand Vilalta, Jonatan Moreno, Eduard Ayguadé, Jesús Labarta, Ulises Cortés, Toyotaro Suzumura T. `Fluid Communities: A Competitive and Highly Scalable Community Detection Algorithm. <https://link.springer.com/chapter/10.1007/978-3-319-72150-7_19/>`_
    """

    g = convert_graph_formats(g_original, nx.Graph)

    coms = nx.algorithms.community.asyn_fluidc(g, k)
    coms = [list(x) for x in coms]
    return NodeClustering(coms, g_original, "Fluid", method_parameters={"k": k})


def der(
    g_original: object,
    walk_len: int = 3,
    threshold: float = 0.00001,
    iter_bound: int = 50,
) -> NodeClustering:
    """
    DER is a Diffusion Entropy Reducer graph clustering algorithm.
    The algorithm uses random walks to embed the graph in a space of measures, after which a modification of k-means in that space is applied. It creates the walks, creates an initialization, runs the algorithm,
    and finally extracts the communities.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: an undirected networkx graph object
    :param walk_len: length of the random walk, default 3
    :param threshold: threshold for stop criteria; if the likelihood_diff is less than threshold tha algorithm stops, default 0.00001
    :param iter_bound: maximum number of iteration, default 50
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.der(G, 3, .00001, 50)


    :References:

    M. Kozdoba and S. Mannor, `Community Detection via Measure Space Embedding <https://papers.nips.cc/paper/5808-community-detection-via-measure-space-embedding/>`_, NIPS 2015

    .. note:: Reference implementation: https://github.com/komarkdev/der_graph_clustering
    """

    graph = convert_graph_formats(g_original, nx.Graph)

    communities, _ = DER.der_graph_clustering(
        graph, walk_len=walk_len, alg_threshold=threshold, alg_iterbound=iter_bound
    )

    maps = {k: v for k, v in enumerate(graph.nodes())}
    coms = []
    for c in communities:
        coms.append([maps[n] for n in c])

    return NodeClustering(
        coms,
        g_original,
        "DER",
        method_parameters={
            "walk_len": walk_len,
            "threshold": threshold,
            "iter_bound": iter_bound,
        },
    )


def frc_fgsn(
    g_original: object, theta: float, eps: float, r: int
) -> FuzzyNodeClustering:
    """Fuzzy-Rough Community Detection on Fuzzy Granular model of Social Network.

    FRC-FGSN assigns nodes to communities specifying the probability of each association.
    The flattened partition ensure that each node is associated to the community that maximize such association probability.
    FRC-FGSN may generate orphan nodes (i.e., nodes not assigned to any community).


    **Supported Graph Types**

    ========== ======== ======== =========
    Undirected Directed Weighted BiPartite
    ========== ======== ======== =========
    Yes        No       No       Yes
    ========== ======== ======== =========

    :param g_original: networkx/igraph object
    :param theta: community density coefficient
    :param eps: coupling coefficient of the community. Ranges in [0, 1], small values ensure that only strongly connected node granules are merged togheter.
    :param r: radius of the granule (int)
    :return: FuzzyNodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = frc_fgsn(G, theta=1, eps=0.5, r=3)


    :References:

    Kundu, S., & Pal, S. K. (2015). `Fuzzy-rough community in social networks. <https://www.sciencedirect.com/science/article/pii/S0167865515000537/>`_ Pattern Recognition Letters, 67, 145-152.

    .. note:: Reference implementation: https://github.com/nidhisridhar/Fuzzy-Community-Detection
    """

    graph = convert_graph_formats(g_original, nx.Graph)
    g, maps = nx_node_integer_mapping(graph)

    communities, fuzz_assoc = fuzzy_comm(graph, theta, eps, r)

    if maps is not None:
        coms = []
        for c in communities:
            coms.append([maps[n] for n in c])

        nx.relabel_nodes(g, maps, False)
        fuzz_assoc = {maps[nid]: v for nid, v in fuzz_assoc.items()}
    else:
        coms = [list(c) for c in communities]

    return FuzzyNodeClustering(
        coms,
        fuzz_assoc,
        g_original,
        "FuzzyComm",
        method_parameters={"theta": theta, "eps": eps, "r": r},
    )


def principled_clustering(
    g_original: object, cluster_count: int
) -> FuzzyNodeClustering:
    """
    An efficient and principled method for detecting communities in networks


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: networkx/igraph object
    :param cluster_count: number of desired communities
    :return: FuzzyNodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = principled_clustering(G, 3)


    :References:

    B Ball, B., & E JNewman, M. (2011). An efficient and principled method for detecting communities in networks. Physical ReviewE, 84, 036103.

    .. note:: Reference implementation: https://github.com/Zabot/principled_clustering
    """

    graph = convert_graph_formats(g_original, nx.Graph)
    g, maps = nx_node_integer_mapping(graph)

    communities, fuzz_assoc = principled(graph, cluster_count)

    if maps is not None:
        coms = []
        for c in communities:
            coms.append([maps[n] for n in c])

        nx.relabel_nodes(g, maps, False)
        fuzz_assoc = {maps[nid]: v for nid, v in fuzz_assoc.items()}
    else:
        coms = [list(c) for c in communities]

    return FuzzyNodeClustering(
        coms,
        fuzz_assoc,
        g_original,
        "Principled Clustering",
        method_parameters={"cluster_count": cluster_count},
    )


def sbm_dl(
    g_original: object,
) -> NodeClustering:
    """Efficient Monte Carlo and greedy heuristic for the inference of stochastic block models.

    Fit a non-overlapping stochastic block model (SBM) by minimizing its description length using an agglomerative heuristic.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: network/igraph object
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = sbm_dl(G)


    :References:

    Tiago P. Peixoto, “Efficient Monte Carlo and greedy heuristic for the inference of stochastic block models”, Phys. Rev. E 89, 012804 (2014), DOI: 10.1103/PhysRevE.89.012804 [sci-hub, @tor], arXiv: 1310.4378.

    .. note:: Implementation from graph-tool library, please report to https://graph-tool.skewed.de for details
    """

    if gt is None:
        raise Exception(
            "===================================================== \n"
            "The graph-tool library seems not to be installed (or incorrectly installed). \n"
            "Please check installation procedure there "
            "https://git.skewed.de/count0/graph-tool/wikis/installation-instructions#native-installation \n"
            "on linux/mac, you can use package managers to do so "
            "(apt-get install python3-graph-tool, brew install graph-tool, etc.)"
        )
    gt_g = convert_graph_formats(g_original, nx.Graph)
    gt_g, label_map = __from_nx_to_graph_tool(gt_g)
    state = gt.minimize_blockmodel_dl(gt_g)

    affiliations = state.get_blocks().get_array()
    affiliations = {label_map[i]: affiliations[i] for i in range(len(affiliations))}
    coms = affiliations2nodesets(affiliations)
    coms = [list(v) for k, v in coms.items()]
    return NodeClustering(coms, g_original, "SBM", method_parameters={})


def sbm_dl_nested(
    g_original: object,
) -> NodeClustering:
    """Efficient Monte Carlo and greedy heuristic for the inference of stochastic block models. (nested)

    Fit a nested non-overlapping stochastic block model (SBM) by minimizing its description length using an agglomerative heuristic.
    Return the lowest level found. Currently cdlib do not support hierarchical clustering.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: igraph/networkx object
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = sbm_dl(G)


    :References:

    Tiago P. Peixoto, “Hierarchical block structures and high-resolution model selection in large networks”, Physical Review X 4.1 (2014): 011047

    .. note:: Implementation from graph-tool library, please report to https://graph-tool.skewed.de for details
    """
    global gt

    if gt is None:
        try:
            import graph_tool.all as gt
        except ModuleNotFoundError:
            raise Exception(
                "===================================================== \n"
                "The graph-tool library seems not to be installed (or incorrectly installed). \n"
                "Please check installation procedure "
                "there https://git.skewed.de/count0/graph-tool/wikis/installation-instructions#native-installation \n"
                "on linux/mac, you can use package managers to do so(apt-get install python3-graph-tool, "
                "brew install graph-tool, etc.)"
            )

    gt_g = convert_graph_formats(g_original, nx.Graph)
    gt_g, label_map = __from_nx_to_graph_tool(gt_g)
    state = gt.minimize_nested_blockmodel_dl(gt_g)

    level0 = state.get_levels()[0]

    affiliations = level0.get_blocks().get_array()
    affiliations = {label_map[i]: affiliations[i] for i in range(len(affiliations))}
    coms = affiliations2nodesets(affiliations)
    coms = [list(v) for k, v in coms.items()]
    return NodeClustering(
        coms,
        g_original,
        "SBM_nested",
        method_parameters={},
    )


def markov_clustering(
    g_original: object,
    expansion: int = 2,
    inflation: int = 2,
    loop_value: int = 1,
    iterations: int = 100,
    pruning_threshold: float = 0.001,
    pruning_frequency: int = 1,
    convergence_check_frequency: int = 1,
) -> NodeClustering:
    """
    The Markov clustering algorithm (MCL) is based on simulation of (stochastic) flow in graphs.
    The MCL algorithm finds cluster structure in graphs by a mathematical bootstrapping procedure. The process deterministically computes (the probabilities of) random walks through the graph, and uses two operators transforming one set of probabilities into another. It does so using the language of stochastic matrices (also called Markov matrices) which capture the mathematical concept of random walks on a graph.
    The MCL algorithm simulates random walks within a graph by alternation of two operators called expansion and inflation.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param expansion: The cluster expansion factor
    :param inflation: The cluster inflation factor
    :param loop_value: Initialization value for self-loops
    :param iterations: Maximum number of iterations
           (actual number of iterations will be less if convergence is reached)
    :param pruning_threshold: Threshold below which matrix elements will be set set to 0
    :param pruning_frequency: Perform pruning every 'pruning_frequency'
           iterations.
    :param convergence_check_frequency: Perform the check for convergence
           every convergence_check_frequency iterations
    :return:  NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.markov_clustering(G)

    :References:

    Enright, Anton J., Stijn Van Dongen, and Christos A. Ouzounis. `An efficient algorithm for large-scale detection of protein families. <https://www.ncbi.nlm.nih.gov/pubmed/11917018/>`_ Nucleic acids research 30.7 (2002): 1575-1584.

    .. note:: Reference implementation: https://github.com/GuyAllard/markov_clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)
    g, maps = nx_node_integer_mapping(g)

    if maps is not None:
        matrix = nx.to_scipy_sparse_matrix(g, nodelist=range(len(maps)))
    else:
        matrix = nx.to_scipy_sparse_matrix(g)

    result = mc.run_mcl(
        matrix,
        expansion=expansion,
        inflation=inflation,
        loop_value=loop_value,
        iterations=iterations,
        pruning_threshold=pruning_threshold,
        pruning_frequency=pruning_frequency,
        convergence_check_frequency=convergence_check_frequency,
    )  # run MCL with default parameters
    clusters = mc.get_clusters(result)

    coms = []
    if maps is not None:
        for c in clusters:
            coms.append([maps[n] for n in c])

        nx.relabel_nodes(g, maps, False)
    else:
        coms = [list(c) for c in clusters]

    return NodeClustering(
        coms,
        g_original,
        "Markov Clustering",
        method_parameters={
            "expansion": expansion,
            "inflation": inflation,
            "loop_value": loop_value,
            "iterations": iterations,
            "pruning_threshold": pruning_threshold,
            "pruning_frequency": pruning_frequency,
            "convergence_check_frequency": convergence_check_frequency,
        },
    )


def chinesewhispers(
    g_original: object, weighting: str = "top", iterations: int = 20, seed: int = None
) -> NodeClustering:
    """

    Fuzzy graph clustering that (i) creates an intermediate representation of the input graph, which reflects the “ambiguity” of its nodes,
    and (ii) uses hard clustering to discover crisp clusters in such “disambiguated” intermediate graph.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original:
    :param weighting: edge weighing schemas. Available modalities: ['top', 'lin', 'log']
    :param iterations: number of iterations
    :param seed: random seed
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.chinesewhispers(G)

    :References:

    Biemann, Chris. 2006. Chinese Whispers: An Efficient Graph Clustering Algorithm and Its Application to Natural Language Processing Problems. In Proceedings of the First Workshop on Graph Based Methods for Natural Language Processing, TextGraphs-1, pages 73–80, Association for Computational Linguistics, New York, NY, USA.

    .. note:: Reference implementation: https://github.com/nlpub/chinese-whispers-python
    """

    g = convert_graph_formats(g_original, nx.Graph)
    g, maps = nx_node_integer_mapping(g)

    cw(g, weighting=weighting, iterations=iterations, seed=seed)

    coms = []
    if maps is not None:
        for _, cluster in sorted(
            aggregate_clusters(g).items(), key=lambda e: len(e[1]), reverse=True
        ):
            coms.append([maps[n] for n in cluster])

        nx.relabel_nodes(g, maps, False)
    else:
        for _, cluster in sorted(
            aggregate_clusters(g).items(), key=lambda e: len(e[1]), reverse=True
        ):
            coms.append(list(cluster))

    return NodeClustering(
        coms,
        g_original,
        "Chinese Whispers",
        method_parameters={"weighting": weighting, "iterations": iterations},
    )


def edmot(
    g_original: object, component_count: int = 2, cutoff: int = 10
) -> NodeClustering:
    """
    The algorithm first creates the graph of higher order motifs. This graph is clustered by the Louvain method.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param component_count: Number of extracted motif hypergraph components. Default is 2.
    :param cutoff: Motif edge cut-off value. Default is 10.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.edmot(G, max_loop=1000)

    :References:

    Li, Pei-Zhen, et al. "EdMot: An Edge Enhancement Approach for Motif-aware Community Detection." Proceedings of the 25th ACM SIGKDD International Conference on Knowledge Discovery & Data Mining. 2019.

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """
    global karateclub

    if "karateclub" not in sys.modules:
        try:
            import karateclub
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install karateclub to use the selected feature."
            )

    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.EdMot(component_count=2, cutoff=10)

    model.fit(g)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(n)

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "EdMot",
        method_parameters={"component_count": component_count, "cutoff": cutoff},
    )


def siblinarity_antichain(
    g_original: object,
    forwards_backwards_on: bool = True,
    backwards_forwards_on: bool = False,
    Lambda: int = 1,
    with_replacement: bool = False,
) -> NodeClustering:
    """
    The algorithm extract communities from a DAG that (i) respects its intrinsic order and (ii) are composed of similar nodes.
    The approach takes inspiration from classic similarity measures of bibliometrics, used to assess how similar two publications are, based on their relative citation patterns.


    **Supported Graph Types**

    ========== ========= ========
    Undirected Directed  Weighted
    ========== ========= ========
    No         Yes (DAG) No
    ========== ========= ========

    :param g_original: a networkx/igraph object representing a DAG (directed acyclic graph)
    :param forwards_backwards_on: checks successors' similarity. Boolean, default True
    :param backwards_forwards_on: checks predecessors' similarity. Boolean, default True
    :param Lambda: desired resolution of the partition. Default 1
    :param with_replacement: If True he similarity of a node to itself is equal to the number of its neighbours based on which the similarity is defined. Boolean, default True.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.siblinarity_antichain(G, Lambda=1)

    :References:

    Vasiliauskaite, V., Evans, T.S. Making communities show respect for order. Appl Netw Sci 5, 15 (2020). https://doi.org/10.1007/s41109-020-00255-5

    .. note:: Reference implementation: https://github.com/vv2246/siblinarity_antichains
    """

    g = convert_graph_formats(g_original, nx.Graph)

    if not nx.is_directed_acyclic_graph(g):
        raise Exception(
            "The Siblinarity Antichain algorithm require as input a Directed Acyclic Graph (DAG)."
        )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        result_list = matrix_node_recursive_antichain_partition(
            g,
            forwards_backwards_on=forwards_backwards_on,
            backwards_forwards_on=backwards_forwards_on,
            Q_check_on=True,
            Lambda=Lambda,
            with_replacement=with_replacement,
            space_label=None,
            time_label=None,
        )

    node_partition = {}
    for n in g.nodes():
        p_at_level = result_list[0]["n_to_p"][n]
        for i in range(1, len(result_list) - 1):
            p_at_level = result_list[i]["n_to_p"][p_at_level]
        node_partition[n] = p_at_level

    partition = defaultdict(list)
    for key, val in node_partition.items():
        partition[val].append(key)

    coms = [list(c) for c in partition.values()]

    return NodeClustering(
        coms,
        g_original,
        "Siblinarity Antichain",
        method_parameters={
            "forwards_backwards_on": forwards_backwards_on,
            "backwards_forwards_on": backwards_forwards_on,
            "Lambda": Lambda,
            "with_replacement": with_replacement,
        },
    )


def ga(
    g_original: object, population: int = 300, generation: int = 30, r: float = 1.5
) -> NodeClustering:
    """
    Genetic based approach to discover communities in social networks.
    GA optimizes a simple but efficacious fitness function able to identify densely connected groups of nodes with sparse connections between groups.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param population:
    :param generation:
    :param r:
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.ga(G)

    :References:

     Pizzuti, C. (2008). Ga-net: A genetic algorithm for community detection in social networks. In Inter conf on parallel problem solving from nature, pages 1081–1090.Springer.

    .. note:: Reference implementation: https://github.com/hariswb/ga-community-detection
    """

    g = convert_graph_formats(g_original, nx.Graph)
    flag = False
    for _, _, d in g.edges(data=True):
        if len(d) > 0:
            flag = True
        d.clear()

    if flag:
        warnings.warn(
            "GA only works on unweighted graphs: edge attributes have been removed from the input network"
        )

    coms = ga_community_detection(g, population, generation, r)

    return NodeClustering(
        coms,
        g_original,
        "ga",
        method_parameters={"population": population, "generation": generation, "r": r},
    )


def belief(
    g_original: object,
    max_it: int = 100,
    eps: float = 0.0001,
    reruns_if_not_conv: int = 5,
    threshold: float = 0.005,
    q_max: int = 7,
) -> NodeClustering:
    """
    Belief community seeks the consensus of many high-modularity partitions.
    It does this with a scalable message-passing algorithm, derived by treating the modularity as a Hamiltonian and applying the cavity method.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param max_it:
    :param eps:
    :param reruns_if_not_conv:
    :param threshold:
    :param q_max:
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.belief(G)

    :References:

    Zhang, Pan, and Cristopher Moore. "Scalable detection of statistically significant communities and hierarchies, using message passing for modularity." Proceedings of the National Academy of Sciences 111.51 (2014): 18144-18149.

    .. note:: Reference implementation: https://github.com/weberfm/belief_propagation_community_detection
    """

    g = convert_graph_formats(g_original, nx.Graph)

    mapping = {n: i for i, n in enumerate(g.nodes())}
    inv_map = {v: k for k, v in mapping.items()}
    g = nx.relabel_nodes(g, mapping)

    coms = detect_belief_communities(
        g,
        max_it=max_it,
        eps=eps,
        reruns_if_not_conv=reruns_if_not_conv,
        threshold=threshold,
        q_max=q_max,
    )

    res = []
    for com in coms:
        com = [inv_map[c] for c in com]
        res.append(com)

    return NodeClustering(
        res,
        g_original,
        "Belief",
        method_parameters={
            "max_it": max_it,
            "eps": eps,
            "reruns_if_not_conv": reruns_if_not_conv,
            "threshold": threshold,
            "q_max": q_max,
        },
    )


def threshold_clustering(
    g_original: object, threshold_function: Callable[[list], float] = np.mean
) -> NodeClustering:
    """
    Developed for semantic similarity networks, this algorithm specifically targets **weighted** and **directed** graphs.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        Yes      Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param threshold_function: callable, optional
        Ties smaller than threshold_function(out_ties) are deleted. Example: np.mean, np.median. Default is np.mean.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.threshold_clustering(G)

    :References:

    Guzzi, Pietro Hiram, Pierangelo Veltri, and Mario Cannataro. "Thresholding of semantic similarity networks using a spectral graph-based technique." International Workshop on New Frontiers in Mining Complex Patterns. Springer, Cham, 2013.

    """

    g = convert_graph_formats(g_original, nx.Graph)

    if not nx.is_directed(g):
        warnings.warn(
            "Threshold Clustering is defined for directed graphs: the undirected graph in input will be treated as directed."
        )

    if not nx.is_weighted(g):
        raise ValueError("Threshold Clustering is defined only for weighted graphs.")

    coms, _ = th_best_partition(g, threshold_function=threshold_function)

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in coms.items():
        coms_to_node[c].append(n)

    coms_louvain = [list(c) for c in coms_to_node.values()]
    return NodeClustering(
        coms_louvain, g_original, "Threshold Clustering", method_parameters={}
    )


def lswl(
    g_original: object,
    query_node: object,
    strength_type: int = 2,
    timeout: float = 1.0,
    online: bool = True,
) -> NodeClustering:
    """

    LSWL locally discovers networks' the communities precisely, deterministically, and quickly.
    This method works in a one-node-expansion model based on a notion of strong and weak links in a graph.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param timeout: The maximum time in which LSWL should retrieve the community. Default is 1 second.
    :param strength_type: 1 strengths between [-1,+1] or, 2 strengths between [0,1]. Default, 2.
    :param query_node: Id of the network node whose local community is queried.
    :param online: wehter the computation should happen in memory or not. Default, True.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.lswl(G, 1)

    :References:

    Fast Local Community Discovery: Relying on the Strength of Links (submitted for KDD 2021).

    .. note:: Reference implementation: https://github.com/mahdi-zafarmand/LSWL

    """

    g = convert_graph_formats(g_original, nx.Graph)
    if online:
        community_searcher = LSWLCommunityDiscovery(g, strength_type, timeout)
    else:
        community_searcher = LSWLCommunityDiscovery_offline(g, strength_type, timeout)

    community = community_searcher.community_search(start_node=query_node)
    community_searcher.reset()

    return NodeClustering(
        [community],
        g_original,
        "LSWL",
        method_parameters={
            "query_node": query_node,
            "strength_type": strength_type,
            "timeout": timeout,
            "online": online,
        },
    )


def lswl_plus(
    g_original: object,
    strength_type: int = 1,
    merge_outliers: bool = True,
    detect_overlap: bool = False,
) -> NodeClustering:
    """
    LSWL+ is capable of finding a partition with overlapping communities or without them, based on user preferences.
    This method can also find outliers (peripheral nodes of the graph that are marginally connected to communities) and hubs (nodes that bridge the communities)


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param strength_type: 1 strengths between [-1,+1] or, 2 strengths between [0,1]. Default, 2.
    :param merge_outliers: If outliers need to merge into communities. Default, True.
    :param detect_overlap: If overlapping communities need to be detected. Default, False
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.lswl_plus(G)

    :References:

    Fast Local Community Discovery: Relying on the Strength of Links (submitted for KDD 2021)

    .. note:: Reference implementation: https://github.com/mahdi-zafarmand/LSWL

    """

    g = convert_graph_formats(g_original, nx.Graph)

    community_detector = LSWLPlusCommunityDetection(
        deepcopy(g), strength_type, merge_outliers, detect_overlap
    )
    partition = community_detector.community_detection()

    return NodeClustering(
        partition,
        g_original,
        "LSWL+",
        method_parameters={
            "strength_type": strength_type,
            "merge_outliers": merge_outliers,
            "detect_overlap": detect_overlap,
        },
    )


def mod_r(g_original: object, query_node: object) -> NodeClustering:
    """
    Community Discovery algorithm that infers the hierarchy of communities that enclose a given vertex by exploring the graph one vertex at a time.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========


    :param g_original: a networkx/igraph object
    :param query_node: Id of the network node whose local community is queried.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.mod_r(G, 1)

    :References:

    Clauset, Aaron. "Finding local community structure in networks." Physical review E 72.2 (2005): 026132.

    .. note:: Reference implementation: https://github.com/mahdi-zafarmand/LSWL

    """

    g = convert_graph_formats(g_original, nx.Graph)
    community_searcher = ModularityRCommunityDiscovery(g)
    community = community_searcher.community_search(start_node=query_node)
    community_searcher.reset()

    return NodeClustering(
        [community], g_original, "mod_r", method_parameters={"query_node": query_node}
    )


def mod_m(g_original: object, query_node: object) -> NodeClustering:
    """
    Community Discovery algorithm designed to find local optimal community structures in large networks starting from a given source vertex.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param query_node: Id of the network node whose local community is queried.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.mod_m(G, 1)

    :References:

    Luo, Feng, James Z. Wang, and Eric Promislow. "Exploring local community structures in large networks." Web Intelligence and Agent Systems: An International Journal 6.4 (2008): 387-400.

    .. note:: Reference implementation: https://github.com/mahdi-zafarmand/LSWL

    """

    g = convert_graph_formats(g_original, nx.Graph)
    community_searcher = ModularityMCommunityDiscovery(g)
    community = community_searcher.community_search(start_node=query_node)
    community_searcher.reset()

    return NodeClustering(
        [community], g_original, "mod_m", method_parameters={"query_node": query_node}
    )


def head_tail(g_original: object, head_tail_ratio: float = 0.4) -> NodeClustering:
    """
    Identifying homogeneous communities in complex networks by applying head/tail breaks on edge betweenness given its heavy-tailed distribution.

    Note: this implementation is suited for small-medium sized graphs, and it may take couple of minutes or longer for a bigger graph.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param head_tail_ratio: head/tail division rule. Float in [0,1], dafault 0.4.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.head_tail()
    >>> coms = algorithms.head_tail(G, head_tail_ratio=0.8)

    :References:

    Jiang B. and Ding M. (2015), Defining least community as a homogeneous group in complex networks, Physica A, 428, 154-160.

    .. note:: Reference implementation: https://github.com/dingmartin/HeadTailCommunityDetection

    """

    g = convert_graph_formats(g_original, nx.Graph)
    coms = HeadTail(g)

    return NodeClustering(
        coms,
        g_original,
        "head_tail",
        method_parameters={"head_tail_ratio": head_tail_ratio},
    )


def kcut(g_original: object, kmax: int = 4) -> NodeClustering:
    """
    An Efficient Spectral Algorithm for Network Community Discovery.
    Kcut is designed to provide a unique combination of recursive partitioning and direct k-way methods, able to guarantee the efficiency of a recursive approach, while also having the same accuracy as a direct k-way method.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param kmax: maximum value of k, dafault 4.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.head_tail()
    >>> coms = algorithms.kcut(G, head_tail_ratio=0.8)

    :References:

    Ruan, Jianhua, and Weixiong Zhang. "An efficient spectral algorithm for network community discovery and its applications to biological and social networks." Seventh IEEE International Conference on Data Mining (ICDM 2007). IEEE, 2007.

    .. note:: Reference implementation: https://github.com/hmliangliang/kcut-algorithm

    """

    g = convert_graph_formats(g_original, nx.Graph)
    coms = kcut_exec(g, kmax)

    return NodeClustering(coms, g_original, "Kcut", method_parameters={"kmax": kmax})


def gemsec(
    g_original: object,
    walk_number: int = 5,
    walk_length: int = 80,
    dimensions: int = 32,
    negative_samples: int = 5,
    window_size: int = 5,
    learning_rate: float = 0.1,
    clusters: int = 10,
    gamma: float = 0.1,
    seed: int = 42,
) -> NodeClustering:
    """
    The procedure uses random walks to approximate the pointwise mutual information matrix obtained by pooling normalized adjacency matrix powers.
    This matrix is decomposed by an approximate factorization technique which is combined with a k-means like clustering cost.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        Yes      No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param walk_number: Number of random walks. Default is 5.
    :param walk_length: Length of random walks. Default is 80.
    :param dimensions: Dimensionality of embedding. Default is 32.
    :param negative_samples: Number of negative samples. Default is 5.
    :param window_size: Matrix power order. Default is 5.
    :param learning_rate: Gradient descent learning rate. Default is 0.1.
    :param clusters: Number of cluster centers. Default is 10.
    :param gamma: Clustering cost weight coefficient. Default is 0.1.
    :param seed: Random seed value. Default is 42.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.gemsec(G)

    :References:

    Rozemberczki, B., Davies, R., Sarkar, R., & Sutton, C. (2019, August). Gemsec: Graph embedding with self clustering. In Proceedings of the 2019 IEEE/ACM international conference on advances in social networks analysis and mining (pp. 65-72).

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """
    global karateclub
    if "karateclub" not in sys.modules:
        try:
            import karateclub
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install karateclub to use the selected feature."
            )

    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.GEMSEC(
        walk_number=walk_number,
        walk_length=walk_length,
        dimensions=dimensions,
        negative_samples=negative_samples,
        window_size=window_size,
        learning_rate=learning_rate,
        clusters=clusters,
        gamma=gamma,
        seed=seed,
    )
    model.fit(g)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(n)

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "GEMSEC",
        method_parameters={
            "walk_number": walk_number,
            "walk_length": walk_length,
            "dimensions": dimensions,
            "negative_samples": negative_samples,
            "window_size": window_size,
            "learning_rate": learning_rate,
            "clusters": clusters,
            "gamma": gamma,
            "seed": seed,
        },
        overlap=False,
    )


def scd(
    g_original: object, iterations: int = 25, eps: float = 1e-06, seed: int = 42
) -> NodeClustering:
    """
    The procedure greedily optimizes the approximate weighted community clustering metric.
    First, clusters are built around highly clustered nodes. Second, we refine the initial partition by using the approximate WCC.
    These refinements happen for the whole vertex set.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param iterations: Refinemeent iterations. Default is 25.
    :param eps: Epsilon score for zero division correction. Default is 10**-6.
    :param seed: Random seed value. Default is 42.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.scd(G)

    :References:

    Prat-Pérez, A., Dominguez-Sal, D., & Larriba-Pey, J. L. (2014, April). High quality, scalable and parallel community detection for large real graphs. In Proceedings of the 23rd international conference on World wide web (pp. 225-236).

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """
    global karateclub

    if "karateclub" not in sys.modules:
        try:
            import karateclub
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install karateclub to use the selected feature."
            )

    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.SCD(iterations=iterations, eps=eps, seed=seed)
    model.fit(g)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(n)

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "SCD",
        method_parameters={"iterations": iterations, "eps": eps, "seed": seed},
        overlap=False,
    )


def pycombo(
    g_original: object,
    weight: str = "weight",
    max_communities: int = None,
    modularity_resolution: float = 1.0,
    num_split_attempts: int = 0,
    start_separate: bool = False,
    treat_as_modularity: bool = False,
    random_seed: int = 42,
) -> NodeClustering:
    """
    This is an implementation (for Modularity maximization) of the community detection algorithm called "Combo".


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param weight: Optional, defaults to weight. Graph edges property to use as weights. If None, graph assumed to be unweighted. Ignored if graph is passed as string (path to the file), or such property does not exist.
    :param max_communities: Optional, defaults to None. Maximum number of communities. If <= 0 or None, assume to be infinite.
    :param modularity_resolution: float, defaults to 1.0. Modularity resolution parameter.
    :param num_split_attempts: int, defaults to 0. Number of split attempts. If 0, autoadjust this number automatically.
    :param start_separate: bool, default False. Indicates if Combo should start from assigning each node into its own separate community. This could help to achieve higher modularity, but it makes execution much slower.
    :param treat_as_modularity:  bool, default False. Indicates if edge weights should be treated as modularity scores. If True, the algorithm solves clique partitioning problem over the given graph, treated as modularity graph (matrix). For example, this allows users to provide their own custom 'modularity' matrix. modularity_resolution is ignored in this case.
    :param random_seed: int, defaults to 42. Random seed to use.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.pycombo(G)

    :References:

    Sobolevsky, S., Campari, R., Belyi, A. and Ratti, C., 2014. General optimization technique for high-quality community detection in complex networks. Physical Review E, 90(1), p.012811.

    .. note:: Reference implementation: https://github.com/Casyfill/pyCombo
    """

    if pycombo_part is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install pycombo to use the selected feature."
        )

    g = convert_graph_formats(g_original, nx.Graph)
    partition = pycombo_part.execute(
        g,
        weight=weight,
        max_communities=max_communities,
        modularity_resolution=modularity_resolution,
        return_modularity=False,
        num_split_attempts=num_split_attempts,
        start_separate=start_separate,
        treat_as_modularity=treat_as_modularity,
        random_seed=random_seed,
    )

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in partition.items():
        coms_to_node[c].append(n)
    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms, g_original, "pyCombo", method_parameters={}, overlap=False
    )


def paris(g_original: object) -> NodeClustering:
    """
    Paris is a hierarchical graph clustering algorithm inspired by modularity-based clustering techniques.
    The algorithm is agglomerative and based on a simple distance between clusters induced by the probability of sampling node pairs.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.paris(G)

    :References:

    Bonald, T., Charpentier, B., Galland, A., & Hollocou, A. (2018). Hierarchical graph clustering using node pair sampling. arXiv preprint arXiv:1806.01664.

    .. note:: Reference implementation: https://github.com/tbonald/paris
    """
    g = convert_graph_formats(g_original, nx.Graph)
    D = paris_alg(g)
    clustering = paris_best_clustering(D)

    return NodeClustering(
        clustering, g_original, "Paris", method_parameters={}, overlap=False
    )


def ricci_community(
    g_original: object, alpha: float = 0.5, method: str = "Sinkhorn"
) -> NodeClustering:
    """
    Curvature is a geometric property to describe the local shape of an object. If we draw two parallel paths on a surface with positive curvature like a sphere, these two paths move closer to each other while for a negatively curved surface like a saddle, these two paths tend to be apart.
    Currently there are multiple ways to discretize curvature on graph, in this algorithm, we include two of the most frequently used discrete Ricci curvature: Ollivier-Ricci curvature which is based on optimal transportation theory and Forman-Ricci curvature which is base on CW complexes.
    Edge Ricci curvature is observed to play an important role in the graph structure.
    An edge with positive curvature represents an edge within a cluster, while a negatively curved edge tent to be a bridge within clusters.
    Also, negatively curved edges are highly related to graph connectivity, with negatively curved edges removed from a connected graph, the graph soon become disconnected.
    Ricci flow is a process to uniformized the edge Ricci curvature of the graph.
    For a given graph, the Ricci flow gives a "Ricci flow metric" on each edge as edge weights, such that under these edge weights, the Ricci curvature of the graph is mostly equal everywhere. In [Ni3], this "Ricci flow metric" is shown to be able to detect communities.
    Both Ricci curvature and Ricci flow metric can act as a graph fingerprint for graph classification.
    The different graph gives different edge Ricci curvature distributions and different Ricci flow metric.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param alpha: The parameter for the probability distribution, range from [0 ~ 1]. It means the share of mass to leave on the original node. Default, 0.5.
    :param method: Transportation method. [“OTD”, “ATD”, “Sinkhorn”]. Default: Sinkhorn
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.ricci_community(G)

    :References:

    Ni, C. C., Lin, Y. Y., Luo, F., & Gao, J. (2019). Community detection on networks with ricci flow. Scientific reports, 9(1), 1-12.

    .. note:: Reference implementation: https://github.com/saibalmars/GraphRicciCurvature
    """
    if OllivierRicci is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install GraphRicciCurvature to use the selected feature."
        )

    g = convert_graph_formats(g_original, nx.Graph)

    cricci = OllivierRicci(g, alpha=alpha, method=method)
    _, clustering = cricci.ricci_community()
    coms = defaultdict(list)

    for k, v in clustering.items():
        coms[v].append(k)

    return NodeClustering(
        list(coms.values()),
        g_original,
        "Ricci",
        method_parameters={"alpha": alpha, "method": method},
        overlap=False,
    )


def spectral(
    g_original: object,
    kmax: int,
    projection_on_smaller_class: bool = True,
    scaler: Callable = None,
) -> NodeClustering:
    """
    SCD implements a Spectral Clustering algorithm for Communities Discovery.
    It is based on Fielder’s vector (obtained from the eigenvector related to the second eigenvalue of the normalized Laplacian) that are leveraged to extract the communities using Kmeans clustering.
    SCD a hierarchical graph clustering algorithm inspired by modularity-based clustering techniques.
    The algorithm is agglomerative and based on a simple distance between clusters induced by the probability of sampling node pairs.


    **Supported Graph Types**

    ========== ======== ======== =========
    Undirected Directed Weighted Bipartite
    ========== ======== ======== =========
    Yes        No       No       Yes
    ========== ======== ======== =========

    :param g_original: a networkx/igraph object
    :param kmax: maximum number of desired communities
    :param projection_on_smaller_class: a boolean value that if True then it project a bipartite network in the smallest class of node. (default is True)
    :param scaler: the function to scale the fielder’s vector to apply KMeans
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.spectral(G)

    :References:

    Higham, Desmond J., Gabriela Kalna, and Milla Kibble. "Spectral clustering and its use in bioinformatics." Journal of computational and applied mathematics 204.1 (2007): 25-37.

    .. note:: Implementation provided by Gianmarco Pepi <g.pepi2@unipi.it>,  Monia Bennici <m.bennici4@studenti.unipi.it>,  Khashayar Abtin <k.abtin@studenti.unipi.it> and Kamran Mehravar <k.mehravar@studenti.unipi.it> (Computer Science Dept., University of Pisa, Italy)
    """

    from sklearn.preprocessing import StandardScaler

    if scaler is None:
        scaler = StandardScaler()

    g = convert_graph_formats(g_original, nx.Graph)
    clustering = spectral_communities(
        g,
        kmax=kmax,
        projection_on_smaller_class=projection_on_smaller_class,
        scaler=scaler,
    )

    return NodeClustering(
        clustering,
        g_original,
        "spectral",
        method_parameters={
            "kmax": kmax,
            "projection_on_smaller_class": projection_on_smaller_class,
        },
        overlap=False,
    )


def mcode(
    g_original: object,
    weights: str = None,
    weight_threshold: float = 0.2,
) -> NodeClustering:
    """
    MCODE is the earliest seed-growth method for predicting protein complexes from PPI networks. MCODE works in two steps:

    1. vertex weighting, and
    2. molecular complex prediction.

    In the vertex weighting step, the weight of a vertex v in the PPI network is calculated from the highest k-core of v’s neighborhood, including v.
    The k-core of a graph is a subgraph where every node is of degree k or greater; the highest k-core is simply the k-core with the highest value of k.
    The weight of v is deﬁned as this maximum k times the density of the corresponding k-core.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param weights: label used for the edge weights. Default, None.
    :param weight_threshold: Threshold for similarity weighs
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.mcode(G)

    :References:

    Bader, G.D., Hogue, C.W. 2003. An automated method for ﬁnding molecular complexes in large protein interaction networks. BMC Bioinformatics 4, 2.

    .. note:: Reference Implementation: https://github.com/trueprice/python-graph-clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)
    clustering = m_code(
        g,
        weights=weights,
        weight_threshold=weight_threshold,
    )

    return NodeClustering(
        clustering,
        g_original,
        "mcode",
        method_parameters={
            "weight_threshold": weight_threshold,
        },
        overlap=False,
    )


def r_spectral_clustering(
    g_original: object,
    n_clusters: int = 2,
    method: str = "regularized",
    percentile: int = None,
) -> NodeClustering:
    """
    Spectral clustering partitions the nodes of a graph into groups based upon the eigenvectors of the graph Laplacian.
    Despite the claims of spectral clustering being “popular”, in applied research using graph data, spectral clustering (without regularization) often returns a partition of the nodes that is uninteresting, typically finding a large cluster that contains most of the data and many smaller clusters, each with only a few nodes.
    This method allows to compute spectral clustering with/withouth different regualarization functions designed to address such a limitation.

    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param n_clusters: How many clusters to look at
    :param method: one among "vanilla", "regularized", "regularized_with_kmeans", "sklearn_spectral_embedding", "sklearn_kmeans", "percentile".
    :param percentile: percentile of the degree distribution to perform regularization. Value in [0, 100]. Mandatory if method="percentile" or "regularized", otherwise None
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.r_spectral_clustering(G, n_clusters=2, method="regularized", percentile=20)

    :References:

    Zhang, Yilin, and Karl Rohe. "Understanding Regularized Spectral Clustering via Graph Conductance." arXiv preprint arXiv:1806.01468 (2018).

    .. note:: Reference Implementation: https://github.com/samialabed/regualirsed-spectral-clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)

    dmap = {n: i for i, n in enumerate(g.nodes)}
    reverse_map = {i: n for n, i in dmap.items()}
    nx.relabel_nodes(g, dmap, False)

    graph_eval = rsc_evaluate_graph(
        g, n_clusters=n_clusters, method=method, percentile=percentile
    )

    clustering = defaultdict(list)
    for idn, v in enumerate(graph_eval):
        clustering[v].append(reverse_map[idn])

    clustering = [c for c in clustering.values()]

    return NodeClustering(
        clustering,
        g_original,
        "regularized_spectral_clustering",
        method_parameters={
            "n_clusters": n_clusters,
            "method": method,
            "percentile": percentile,
        },
        overlap=False,
    )



import random

"""
Lancichinetti, Andrea, Santo Fortunato, and János Kertész. 
"Detecting the overlapping and hierarchical algorithms structure in complex networks"
New Journal of Physics 11.3 (2009): 033015.>>
"""


class Community(object):
    def __init__(self, G, alpha=1.0):
        self.g = G
        self.alpha = alpha
        self.nodes = set()
        self.k_in = 0
        self.k_out = 0

    def add_node(self, node):
        neighbors = set(self.g.neighbors(node))
        node_k_in = len(neighbors & self.nodes)
        node_k_out = len(neighbors) - node_k_in
        self.nodes.add(node)
        self.k_in += 2 * node_k_in
        self.k_out = self.k_out + node_k_out - node_k_in

    def remove_vertex(self, node):
        neighbors = set(self.g.neighbors(node))
        community_nodes = self.nodes
        node_k_in = len(neighbors & community_nodes)
        node_k_out = len(neighbors) - node_k_in
        self.nodes.remove(node)
        self.k_in -= 2 * node_k_in
        self.k_out = self.k_out - node_k_out + node_k_in

    def cal_add_fitness(self, node):
        neighbors = set(self.g.neighbors(node))
        old_k_in = self.k_in
        old_k_out = self.k_out
        vertex_k_in = len(neighbors & self.nodes)
        vertex_k_out = len(neighbors) - vertex_k_in
        new_k_in = old_k_in + 2 * vertex_k_in
        new_k_out = old_k_out + vertex_k_out - vertex_k_in
        new_fitness = new_k_in / (new_k_in + new_k_out) ** self.alpha
        old_fitness = old_k_in / (old_k_in + old_k_out) ** self.alpha
        return new_fitness - old_fitness

    def cal_remove_fitness(self, node):
        neighbors = set(self.g.neighbors(node))
        new_k_in = self.k_in
        new_k_out = self.k_out
        node_k_in = len(neighbors & self.nodes)
        node_k_out = len(neighbors) - node_k_in
        old_k_in = new_k_in - 2 * node_k_in
        old_k_out = new_k_out - node_k_out + node_k_in
        old_fitness = old_k_in / (old_k_in + old_k_out) ** self.alpha
        new_fitness = new_k_in / (new_k_in + new_k_out) ** self.alpha
        return new_fitness - old_fitness

    def recalculate(self):
        for vid in self.nodes:
            fitness = self.cal_remove_fitness(vid)
            if fitness < 0.0:
                return vid
        return None

    def get_neighbors(self):
        neighbors = set()
        for node in self.nodes:
            neighbors.update(set(self.g.neighbors(node)) - self.nodes)
        return neighbors

    def get_fitness(self):
        return float(self.k_in) / ((self.k_in + self.k_out) ** self.alpha)


class LFM_nx(object):
    def __init__(self, G, alpha):
        self.g = G
        self.alpha = alpha

    def execute(self):
        communities = []
        node_not_include = list(self.g.nodes.keys())[:]
        while len(node_not_include) != 0:
            c = Community(self.g, self.alpha)
            # randomly select a seed node
            seed = random.choice(node_not_include)
            c.add_node(seed)

            to_be_examined = c.get_neighbors()
            while to_be_examined:
                # largest fitness to be added
                m = {}
                for node in to_be_examined:
                    fitness = c.cal_add_fitness(node)
                    m[node] = fitness
                to_be_add = sorted(m.items(), key=lambda x: x[1], reverse=True)[0]

                # stop condition
                if to_be_add[1] < 0.0:
                    break
                c.add_node(to_be_add[0])

                to_be_remove = c.recalculate()
                while to_be_remove is not None:
                    c.remove_vertex(to_be_remove)
                    to_be_remove = c.recalculate()

                to_be_examined = c.get_neighbors()

            for node in c.nodes:
                if node in node_not_include:
                    node_not_include.remove(node)
            communities.append(list(c.nodes))
        return list(communities)



#!/usr/bin/env python
# encoding:UTF-8

################################################################################################
#
#    Community Detection via Local Spectral Clustering
#
################################################################################################

# (Our algorithm is also known as "LEMON", which is the short form of Local Expansion via Minimum One Norm)


# LEMON.py
# Yixuan Li
# Last modified: 2015-1-8
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.


import numpy as np
import math
import pulp
from scipy import linalg as splin
import gc


def __set_initial_prob(n, starting_nodes):
    """Precondition: starting_nodes is ndarray which indicate the indices of starting points

    Return: A probability vector with n elements
    """
    v = np.zeros(n)
    v[starting_nodes] = 1.0 / starting_nodes.size

    return v


def __set_initial_prob_proportional(n, degree_sequence, starting_nodes):
    """Precondition: starting_nodes is ndarray which indicate the indices of starting points

    Return: A probability vector with n elements
    """
    v = np.zeros(n)
    vol = 0
    for node in starting_nodes:
        vol += degree_sequence[node]
    for node in starting_nodes:
        v[node] = degree_sequence[node] / float(vol)

    return v


def __adj_to_Laplacian(G):
    """Computes the normalized adjacency matrix of a given graph"""

    n = G.shape[0]
    D = np.zeros((1, n))
    for i in range(n):
        D[0, i] = math.sqrt(G[i, :].sum())

    temp = np.dot(D.T, np.ones((1, n)))
    horizontal = G / temp
    normalized_adjacency_matrix = horizontal / (temp.T)
    gc.collect()

    return normalized_adjacency_matrix


def __cal_conductance(G, cluster):
    """cluster: a list of node id that forms a algorithms. Data type of cluster is given by numpy array

    Calculate the conductance of the cut A and complement of A.
    """

    assert (
        type(cluster) == np.ndarray
    ), "The given algorithms members is not a numpy array"

    temp = G[cluster, :]
    subgraph = temp[:, cluster]
    cutsize = temp.sum() - subgraph.sum()
    denominator = min(temp.sum(), G.sum() - temp.sum())
    conductance = cutsize / denominator if denominator > 0 else 1

    return conductance


def __random_walk(G, initial_prob, subspace_dim=3, walk_steps=3):
    """
    Start a random walk with probability distribution p_initial.
    Transition matrix needs to be calculated according to adjacent matrix G.

    """
    assert (
        type(initial_prob) == np.ndarray
    ), "Initial probability distribution is \
                                             not a numpy array"

    # Transform the adjacent matrix to a laplacian matrix P
    P = __adj_to_Laplacian(G)

    Prob_Matrix = np.zeros((G.shape[0], subspace_dim))
    Prob_Matrix[:, 0] = initial_prob
    for i in range(1, subspace_dim):
        Prob_Matrix[:, i] = np.dot(Prob_Matrix[:, i - 1], P)

    Orth_Prob_Matrix = splin.orth(Prob_Matrix)

    for i in range(walk_steps):
        temp = np.dot(Orth_Prob_Matrix.T, P)
        Orth_Prob_Matrix = splin.orth(temp.T)

    return Orth_Prob_Matrix


def __min_one_norm(B, initial_seed, seed):
    weight_initial = 1 / float(len(initial_seed))
    weight_later_added = weight_initial / float(0.5)
    difference = len(seed) - len(initial_seed)
    [r, c] = B.shape
    prob = pulp.LpProblem("Minimum one norm", pulp.LpMinimize)
    indices_y = range(0, r)
    y = pulp.LpVariable.dicts("y_s", indices_y, 0)
    indices_x = range(0, c)
    x = pulp.LpVariable.dicts("x_s", indices_x)

    f = dict(zip(indices_y, [1.0] * r))

    prob += pulp.lpSum(f[i] * y[i] for i in indices_y)  # objective function

    prob += pulp.lpSum(y[s] for s in initial_seed) >= 1

    prob += pulp.lpSum(y[r] for r in seed) >= 1 + weight_later_added * difference

    for j in range(r):
        temp = dict(zip(indices_x, list(B[j, :])))
        prob += pulp.lpSum(y[j] + (temp[k] * x[k] for k in indices_x)) == 0

    prob.solve()

    result = []
    for var in indices_y:
        result.append(y[var].value())

    return result


def __global_minimum(sequence, start_index):
    detected_size = len(list(sequence))
    seq_length = len(list(sequence))
    cond = sequence[seq_length - 2]
    for x in range(40):
        list(sequence).append(0)
    for i in range(seq_length - 40):
        if sequence[i] < sequence[i - 1] and sequence[i] < sequence[i + 1]:
            count_larger = 0
            count_smaller = 0
            for j in range(1, 32):
                if sequence[i + 1 + j] > sequence[i + 1]:
                    count_larger += 1
            for k in range(1, 32):
                if sequence[i - 1 - k] > sequence[i - 1]:
                    count_smaller += 1
            if count_larger >= 18 and count_smaller >= 18:
                detected_size = i + start_index
                cond = sequence[i]
                break
    return detected_size, cond


def lemon(
    G,
    seedset,
    min_comm_size,
    max_comm_size,
    expand_step=None,
    subspace_dim=None,
    walk_steps=None,
    biased=True,
):
    degree = []
    n = G.shape[0]
    for x in range(n):
        degree.append(G[x].sum())

    # Random walk starting from seed nodes:
    if biased:
        initial_prob = __set_initial_prob_proportional(n, degree, seedset)
    else:
        initial_prob = __set_initial_prob(G.shape[0], seedset)

    Orth_Prob_Matrix = __random_walk(G, initial_prob, subspace_dim, walk_steps)
    initial_seed = seedset

    # Initialization
    detected = list(seedset)
    seed = seedset
    step = expand_step
    detected_comm = []

    global_conductance = np.zeros(30)
    global_conductance[-1] = 1000000  # set the last element to be infinitely large
    global_conductance[-2] = 1000000
    flag = True
    iteration = 0

    while iteration < 30 and flag:
        temp = np.argsort(
            np.array(__min_one_norm(Orth_Prob_Matrix, list(initial_seed), list(seed)))
        )

        sorted_top = list(temp[::-1][:step])

        detected = list(set(list(detected) + sorted_top))
        seed = np.array(detected)

        conductance_record = np.zeros(max_comm_size - min_comm_size + 1)
        conductance_record[-1] = 0

        for i in range(min_comm_size, max_comm_size):
            candidate_comm = np.array(list(temp[::-1][:i]))
            conductance_record[i - min_comm_size] = __cal_conductance(G, candidate_comm)

        detected_size, cond = __global_minimum(conductance_record, min_comm_size)

        step += expand_step

        if biased:
            initial_prob = __set_initial_prob_proportional(n, degree, seedset)
        else:
            initial_prob = __set_initial_prob(G.shape[0], seedset)

        Orth_Prob_Matrix = __random_walk(G, initial_prob, subspace_dim, walk_steps)

        if detected_size != 0:
            current_comm = list(temp[::-1][:detected_size])
            detected_comm = current_comm

        global_conductance[iteration] = cond
        if (
            global_conductance[iteration - 1] <= global_conductance[iteration]
            and global_conductance[iteration - 1] <= global_conductance[iteration - 2]
        ):
            flag = False

        iteration += 1

    return detected_comm



import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import random
from queue import Queue
from collections import defaultdict

# from dgl.data import LegacyTUDataset


class EBGC:
    def __init__(self):
        self.g = None
        self.adj = None
        self.node_num = None
        self.node_state = None
        self.node_degree = None
        self.cluster_node = None
        pass

    def node_neighbor_idxs(self, node_idx):
        neigh_list = np.array([n for n in self.g.neighbors(node_idx)])
        # neigh_list = np.array(list(dict.keys(dict(self.g[node_idx]))))
        return neigh_list

    def node_entropy(self, node_idx, cluster_idx):
        neigh_list = self.node_neighbor_idxs(node_idx)
        cluster_list = self.cluster_node[cluster_idx]
        n = np.sum(cluster_list[neigh_list])
        p_inner = n / len(neigh_list)
        p_outer = 1 - p_inner
        if p_inner == 0 or p_inner == 1:
            e_v = 0
        else:
            e_v = -p_inner * np.log2(p_inner) - p_outer * np.log2(p_outer)

        return e_v

    def graph_entropy(self, cluster_idx):
        e_g = 0
        for idx in range(self.node_num):
            e_g += self.node_entropy(idx, cluster_idx)

        return e_g

    def add_node_to_cluster(self, node_idx, cluster_idx):
        self.cluster_node[cluster_idx][node_idx] = 1
        self.node_state[node_idx] = 1

    def rm_node_to_cluster(self, node_idx, cluster_idx):
        self.cluster_node[cluster_idx][node_idx] = 0
        self.node_state[node_idx] = 0

    def cluster_add_neighbors(self, cluster_idx, seed_neighbors):
        cluster_list = self.cluster_node[cluster_idx]
        seed_neighbors = seed_neighbors[np.where(self.node_state[seed_neighbors] == 0)]
        cluster_list[seed_neighbors] = 1
        self.cluster_node[cluster_idx] = cluster_list
        self.node_state[seed_neighbors] = 1
        return seed_neighbors

    def cluster_removal_neighbors(self, cluster_idx, seed_neighbors):

        checked_seed_neighbors = []
        min_e_g = self.graph_entropy(cluster_idx)

        for neigh_idx in seed_neighbors:
            # remove neigh_idx node
            self.rm_node_to_cluster(neigh_idx, cluster_idx)
            e_g = self.graph_entropy(cluster_idx)
            if e_g < min_e_g:
                min_e_g = e_g
            else:
                # add neigh_idx node
                self.add_node_to_cluster(neigh_idx, cluster_idx)
                checked_seed_neighbors.append(neigh_idx)
        return checked_seed_neighbors

    def select_seed_ver(self):
        # Select a seed: Here we follow the decrease order of nodes' degree
        # t = np.argsort(self.node_degree)
        # deg = self.node_degree
        for idx in np.argsort(self.node_degree)[::-1]:
            if self.node_state[idx] == 0:
                # add seed into the cluster (create a cluster)
                seed_idx = idx
                cluster_list = np.zeros(self.node_num)
                self.cluster_node.append(cluster_list)
                cluster_idx = len(self.cluster_node) - 1
                self.add_node_to_cluster(seed_idx, cluster_idx)
                break

        seed_neighbors = self.node_neighbor_idxs(seed_idx)
        seed_neighbors = self.cluster_add_neighbors(cluster_idx, seed_neighbors)

        return cluster_idx, seed_neighbors

    def add_boundary_ver(self, check_list, cluster_idx):
        node_que = Queue()
        [node_que.put(x) for x in check_list]
        min_e_g = self.graph_entropy(cluster_idx)

        while not node_que.empty():
            node_idx = node_que.get()

            # find neighbor of node_idx
            neigh_idx = self.node_neighbor_idxs(node_idx)

            # check neighbor
            # the state of neighbor is 0
            neigh_idx = neigh_idx[np.where(self.node_state[neigh_idx] == 0)]
            # add the neighbor into the cluster which lowers the graph entropy
            for idx in neigh_idx:
                self.add_node_to_cluster(idx, cluster_idx)
                e_g = self.graph_entropy(cluster_idx)
                if e_g < min_e_g:
                    min_e_g = e_g
                    node_que.put(idx)
                else:
                    self.rm_node_to_cluster(idx, cluster_idx)

    def converge_check(self):
        return np.any(self.node_state == 0)

    def fit(self, g):
        self.g = g
        degree = np.array(g.degree)
        degree = degree[np.argsort(degree[:, 0])]
        self.node_degree = [x[1] for x in degree]
        self.node_num = len(self.node_degree)
        self.node_state = np.zeros(self.node_num)
        self.cluster_node = []

        while self.converge_check():

            # Step 1
            cluster_idx, seed_neighbors = self.select_seed_ver()

            # Step 2
            checked_seed_neighbors = self.cluster_removal_neighbors(
                cluster_idx, seed_neighbors
            )

            # Step 3
            self.add_boundary_ver(checked_seed_neighbors, cluster_idx)

        return np.array(self.cluster_node).T


def draw_graph(g, node_labels):
    # pos = nx.spring_layout(g)
    # nx.draw_spring(g, with_labels=False)
    # nx.draw_networkx_edge_labels(g, pos, font_size=14, alpha=0.5, rotate=True)

    k = 1 / (np.max(node_labels))
    val_map = {x: x * k for x in node_labels}
    values = [val_map.get(node_labels[node]) for node in g.nodes()]

    nx.draw(
        g,
        cmap=plt.get_cmap("viridis"),
        node_color=values,
        with_labels=True,
        font_color="white",
    )

    plt.axis("off")
    plt.show()



from cdlib import EdgeClustering
from collections import defaultdict
import networkx as nx
from cdlib.utils import convert_graph_formats
from cdlib.algorithms.internal.HLC import HLC, HLC_read_edge_list_unweighted

__all__ = ["hierarchical_link_community"]


def hierarchical_link_community(g_original: object) -> EdgeClustering:
    """
    HLC (hierarchical link clustering) is a method to classify links into topologically related groups.
    The algorithm uses a similarity between links to build a dendrogram where each leaf is a link from the original network and branches represent link communities.
    At each level of the link dendrogram is calculated the partition density function, based on link density inside communities, to pick the best level to cut.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: EdgeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.hierarchical_link_community(G)

    :References:

    Ahn, Yong-Yeol, James P. Bagrow, and Sune Lehmann. `Link communities reveal multiscale complexity in networks. <https://www.nature.com/articles/nature09182/>`_ nature 466.7307 (2010): 761.
    """

    g = convert_graph_formats(g_original, nx.Graph)

    adj, edges = HLC_read_edge_list_unweighted(g)

    edge2cid, _, _, _ = HLC(adj, edges).single_linkage()

    coms = defaultdict(list)
    for e, com in edge2cid.items():
        coms[com].append(e)

    coms = [list(c) for c in coms.values()]
    return EdgeClustering(coms, g_original, "HLC", method_parameters={})



import networkx as nx
import copy
import random
import numpy as np

__all__ = ["XMark_benchmark"]


def zeta(x, q, tolerance):
    """The Hurwitz zeta function, or the Riemann zeta function of two
    arguments.
    ``x`` must be greater than one and ``q`` must be positive.
    This function repeatedly computes subsequent partial sums until
    convergence, as decided by ``tolerance``.
    """
    try:
        from scipy.special import zeta as _zeta

        return _zeta(x, q)
    except ImportError:
        z = 0
        z_prev = -float("inf")
        k = 0
        while abs(z - z_prev) > tolerance:
            z_prev = z
            z += 1 / ((k + q) ** x)
            k += 1
        return z


def _zipf_rv_below(gamma, xmin, threshold, seed):
    """Returns a random value chosen from the bounded Zipf distribution.
    Repeatedly draws values from the Zipf distribution until the
    threshold is met, then returns that value.
    """
    result = nx.utils.zipf_rv(gamma, xmin, seed)
    while result > threshold:
        result = nx.utils.zipf_rv(gamma, xmin, seed)
    return result


def _powerlaw_sequence(gamma, low, high, condition, length, max_iters, seed):
    """Returns a list of numbers obeying a constrained power law distribution.
    ``gamma`` and ``low`` are the parameters for the Zipf distribution.
    ``high`` is the maximum allowed value for values draw from the Zipf
    distribution. For more information, see :func:`_zipf_rv_below`.
    ``condition`` and ``length`` are Boolean-valued functions on
    lists. While generating the list, random values are drawn and
    appended to the list until ``length`` is satisfied by the created
    list. Once ``condition`` is satisfied, the sequence generated in
    this way is returned.
    ``max_iters`` indicates the number of times to generate a list
    satisfying ``length``. If the number of iterations exceeds this
    value, :exc:`~networkx.exception.ExceededMaxIterations` is raised.
    seed : integer, random_state, or None (default)
        Indicator of random number generation state.
        See :ref:`Randomness<randomness>`.
    """
    for i in range(max_iters):
        seq = []
        while not length(seq):
            seq.append(_zipf_rv_below(gamma, low, high, seed))
        if condition(seq):
            return seq
    raise nx.ExceededMaxIterations("Could not create power law sequence")


def _generate_min_degree(gamma, average_degree, max_degree, tolerance, max_iters):
    """Returns a minimum degree from the given average degree."""
    min_deg_top = max_degree
    min_deg_bot = 1
    min_deg_mid = (min_deg_top - min_deg_bot) / 2 + min_deg_bot
    itrs = 0
    mid_avg_deg = 0
    while abs(mid_avg_deg - average_degree) > tolerance:
        if itrs > max_iters:
            raise nx.ExceededMaxIterations("Could not match average_degree")
        mid_avg_deg = 0
        for x in range(int(min_deg_mid), max_degree + 1):
            mid_avg_deg += (x ** (-gamma + 1)) / zeta(gamma, min_deg_mid, tolerance)
        if mid_avg_deg > average_degree:
            min_deg_top = min_deg_mid
            min_deg_mid = (min_deg_top - min_deg_bot) / 2 + min_deg_bot
        else:
            min_deg_bot = min_deg_mid
            min_deg_mid = (min_deg_top - min_deg_bot) / 2 + min_deg_bot
        itrs += 1
    return round(min_deg_mid)


def _assign_random_labels(seq, labels, lab_imb):
    """For Categorical attributes: assign the purest label of each community.
    Return a list of the purest labels."""
    tot_lab_seq = []
    for lb in labels:
        if lb == "auto":
            card_lab = [i for i in range(1, len(seq) + 1)]
        else:
            card_lab = [i for i in range(1, lb + 1)]
        lab_seq = []
        for i in range(len(seq)):
            if lb == "auto":
                lab_seq.append(i + 1)
            else:
                lab_seq.append(random.randrange(1, lb + 1))

        tot_lab_seq.append(lab_seq)
    return tot_lab_seq


def _assign_random_means(seq, labels, lab_imb, mu):
    """For Categorical attributes: assign the desired mean of each community.
    Return a list of the means."""
    tot_lab_seq = []
    for k, lb in enumerate(labels):
        if lb == "auto":
            dist = [i * 10 for i in range(len(seq))]
            multimodal = [
                random.uniform(dist[i] - 2, dist[i] + 2) for i in range(len(seq))
            ]
        else:
            dist = [i * 10 for i in range(lb)]
            multimodal = [random.uniform(dist[i] - 2, dist[i] + 2) for i in range(lb)]
        lab_seq = []
        for i in range(len(seq)):
            if lb == "auto":
                lab_seq.append(multimodal[i])
            else:
                lab_seq.append(
                    random.choice(multimodal)
                )  # will be the mean of community

        tot_lab_seq.append(lab_seq)
    return tot_lab_seq


def _assign_node_memberships(degree_seq):
    overlap_seq = [2 for el in degree_seq]
    return overlap_seq


def _generate_overlapping_communities(
    degree_seq, community_sizes, overlap_seq, mu, max_iters
):
    result = [set() for _ in community_sizes]
    n = len(degree_seq)
    free = list(range(n))
    for i in range(max_iters):
        v = free.pop()
        for _ in range(overlap_seq[v]):
            c = random.choice(range(len(community_sizes)))
            s = round(degree_seq[v] * (1 - mu))
            # If the community is large enough, add the node to the chosen
            # community. Otherwise, return it to the list of unaffiliated
            # nodes
            if s < community_sizes[c]:
                result[c].add(v)
            else:
                free.append(v)
            # If the community is too big, remove a node from it.
            if len(result[c]) > community_sizes[c]:
                free.append(result[c].pop())


def _generate_communities(
    degree_seq,
    community_sizes,
    lab_coms,
    mu,
    labels,
    noise,
    std,
    max_iters,
    seed,
    type_attr,
):
    """Returns a list of sets, each of which represents a community.
    ``degree_seq`` is the degree sequence that must be met by the
    graph.
    ``community_sizes`` is the community size distribution that must be
    met by the generated list of sets.
    ``mu`` is a float in the interval [0, 1] indicating the fraction of
    intra-community edges incident to each node.
    ``max_iters`` is the number of times to try to add a node to a
    community. This must be greater than the length of
    ``degree_seq``, otherwise this function will always fail. If
    the number of iterations exceeds this value,
    :exc:`~networkx.exception.ExceededMaxIterations` is raised.
    seed : integer, random_state, or None (default)
        Indicator of random number generation state.
        See :ref:`Randomness<randomness>`.
    The communities returned by this are sets of integers in the set {0,
    ..., *n* - 1}, where *n* is the length of ``degree_seq``.
    """
    card_lab = []
    for lb in labels:
        if lb == "auto":  # number of labels equal to the number of communities
            card_lab.append([i for i in range(1, len(community_sizes) + 1)])
        else:
            card_lab.append([i for i in range(1, lb + 1)])

    # This assumes the nodes in the graph will be natural numbers.
    result = [set() for _ in community_sizes]
    n = len(degree_seq)
    lab_nodes = [list(range(n)) for i in range(len(card_lab))]
    free = list(range(n))
    for i in range(max_iters):
        v = free.pop()
        c = random.choice(range(len(community_sizes)))
        s = round(degree_seq[v] * (1 - mu))
        # If the community is large enough, add the node to the chosen
        # community. Otherwise, return it to the list of unaffiliated nodes.
        if s < community_sizes[c]:
            result[c].add(v)

            if type_attr == "continuous":
                for j, attr in enumerate(lab_coms):
                    lab_nodes[j][v] = float(np.random.normal(attr[c], std, 1))
            # if categorical
            else:
                for j, attr in enumerate(lab_coms):
                    if random.uniform(0, 1) < 1 - noise:
                        lab_nodes[j][v] = attr[c]
                    else:
                        l = random.choice(card_lab[j])
                        lab_nodes[j][v] = l
        else:
            free.append(v)
        # If the community is too big, remove a node from it.
        if len(result[c]) > community_sizes[c]:
            free.append(result[c].pop())
        if not free:
            return lab_nodes, result
    msg = "Could not assign communities"
    raise nx.ExceededMaxIterations(msg)


def XMark_benchmark(
    n,
    tau1,
    tau2,
    mu,
    labels=2,
    std=0.1,
    noise=0,
    lab_imb=0,
    average_degree=None,
    min_degree=None,
    max_degree=None,
    min_community=None,
    max_community=None,
    tol=1.0e-7,
    max_iters=500,
    seed=None,
    type_attr="categorical",
):

    # Perform some basic parameter validation.
    if not tau1 > 1:
        raise nx.NetworkXError("tau1 must be greater than one")
    if not tau2 > 1:
        raise nx.NetworkXError("tau2 must be greater than one")
    if not 0 <= mu <= 1:
        raise nx.NetworkXError("mu must be in the interval [0, 1]")

    # Validate parameters for generating the degree sequence.
    if max_degree is None:
        max_degree = n
    elif not 0 < max_degree <= n:
        raise nx.NetworkXError("max_degree must be in the interval (0, n]")
    if not ((min_degree is None) ^ (average_degree is None)):
        raise nx.NetworkXError(
            "Must assign exactly one of min_degree and" " average_degree"
        )
    if min_degree is None:
        min_degree = _generate_min_degree(
            tau1, average_degree, max_degree, tol, max_iters
        )

    # Generate a degree sequence with a power law distribution.
    low, high = min_degree, max_degree

    def condition(seq):
        return sum(seq) % 2 == 0

    def length(seq):
        return len(seq) >= n

    deg_seq = _powerlaw_sequence(tau1, low, high, condition, length, max_iters, seed)

    # Validate parameters for generating the community size sequence.
    if min_community is None:
        min_community = min(deg_seq)
    if max_community is None:
        max_community = max(deg_seq)

    # Generate a community size sequence with a power law distribution.
    low, high = min_community, max_community

    def condition(seq):
        return sum(seq) == n

    def length(seq):
        return sum(seq) >= n

    comms = _powerlaw_sequence(tau2, low, high, condition, length, max_iters, seed)

    overlap_seq = _assign_node_memberships(deg_seq)

    overlap_comms = copy.copy(comms)
    overlap_deg_seq = copy.copy(deg_seq)
    cycle = [overlap_comms, overlap_deg_seq]
    if sum(overlap_comms) < sum(overlap_seq):
        add_size = abs(sum(overlap_comms) - sum(overlap_seq))
        for new in range(add_size):
            for cy in cycle:
                pick = random.choice(list(range(len(cy))))
            if cy[pick] < np.median(cy):
                cy[pick] += 1
            else:
                cy[random.choice(list(range(len(cy))))] += 1

    if type_attr == "continuous":
        lab_coms = _assign_random_means(comms, labels, lab_imb, mu)
    else:
        lab_coms = _assign_random_labels(comms, labels, lab_imb)

    # Generate the communities based on the given degree sequence and
    # community sizes.
    max_iters *= 10 * n

    lab_nodes, communities = _generate_communities(
        deg_seq, comms, lab_coms, mu, labels, noise, std, max_iters, seed, type_attr
    )

    # Finally, generate the benchmark graph based on the given
    # communities, joining nodes according to the intra- and
    # inter-community degrees.
    G = nx.Graph()
    G.add_nodes_from(range(n))

    for i, c in enumerate(communities):

        for u in c:
            while G.degree(u) < round(deg_seq[u] * (1 - mu)):
                v = random.choice(list(c))
                G.add_edge(u, v)

            while G.degree(u) < deg_seq[u]:
                v = random.choice(range(n))
                if v not in c:
                    G.add_edge(u, v)

            for j, lab in enumerate(lab_nodes):
                G.nodes[u]["label_" + str(j)] = lab[u]
            G.nodes[u]["community"] = c

    return G



import networkx as nx

import time
from random import random, shuffle


class ModularityMCommunityDiscovery:
    def __init__(self, graph):
        self.graph = graph
        self.starting_node = None
        self.community = []
        self.boundary = set()
        self.shell = set()
        self.remove_self_loops()

    def reset(self):
        self.community.clear()
        self.boundary.clear()
        self.shell.clear()

    def remove_self_loops(self):
        for node in self.graph.nodes():
            if self.graph.has_edge(node, node):
                self.graph.remove_edge(node, node)

    def set_start_node(self, start_node):
        if start_node in self.graph.nodes():
            self.starting_node = start_node
            self.community.append(start_node)
            self.boundary.add(start_node)
            self.shell = set(self.graph.neighbors(start_node))
        else:
            print("Invalid starting node! Try with another one.")
            exit(-1)

    def update_sets_when_node_joins(self, node, change_boundary=False):
        self.community.append(node)
        if change_boundary:
            self.update_boundary_when_node_joins(node)
        self.update_shell_when_node_joins(node)

    def update_shell_when_node_joins(self, new_node):
        self.shell.update(self.graph.neighbors(new_node))
        for node in self.community:
            self.shell.discard(node)

    def update_boundary_when_node_joins(self, new_node):
        should_be_boundary = False
        for neighbor in self.graph.neighbors(new_node):
            if (neighbor in self.community) is False:
                should_be_boundary = True
                break
        if should_be_boundary:
            self.boundary.add(new_node)

    def update_sets_when_node_leaves(self, node, change_boundary=False):
        self.community.remove(node)
        if change_boundary:
            self.update_boundary_when_node_leaves(node)
        self.update_shell_when_node_leaves(node)

    def update_boundary_when_node_leaves(self, old_node):
        if old_node in self.boundary:
            self.boundary.remove(old_node)
            for node in self.graph.neighbors(old_node):
                if node in self.community:
                    self.boundary.add(node)

    def update_shell_when_node_leaves(self, old_node):
        possibles_leaving_nodes = [
            node for node in self.graph.neighbors(old_node) if node in self.shell
        ]
        for node in possibles_leaving_nodes:
            should_leave_shell = True
            for neighbor in self.graph.neighbors(node):
                if neighbor in self.community:
                    should_leave_shell = False
                    break
            if should_leave_shell:
                self.shell.remove(node)
        self.shell.add(old_node)

    def community_search(self, start_node, with_amend=False):
        self.set_start_node(start_node)
        sorted_shell = list(self.shell)

        modularity = 0.0
        while (
            len(self.community) < self.graph.number_of_nodes() and len(self.shell) > 1
        ):
            Q_list = []
            sorted_shell.sort(key=self.graph.degree)
            for candidate_node in sorted_shell:
                new_modularity = self.compute_modularity("addition", candidate_node)
                if new_modularity > modularity:
                    modularity = new_modularity
                    self.update_sets_when_node_joins(candidate_node)
                    sorted_shell.remove(candidate_node)
                    Q_list.append(candidate_node)

            while True:
                Q_delete = []
                for candidate_node in sorted(self.community, key=lambda x: random()):
                    new_modularity = self.compute_modularity("deletion", candidate_node)
                    if new_modularity > modularity:
                        modularity = new_modularity
                        self.update_sets_when_node_leaves(candidate_node)
                        Q_delete.append(candidate_node)

                        if candidate_node in Q_list:
                            Q_list.remove(candidate_node)

                if len(Q_delete) == 0:
                    break

            for node in sorted(Q_list, key=lambda x: random()):
                neighbors = list(self.graph.neighbors(node))
                shuffle(neighbors)
                for neighbor in neighbors:
                    if (neighbor in self.community) is False:
                        self.shell.add(neighbor)
                        if (neighbor in sorted_shell) is False:
                            sorted_shell.append(neighbor)

            if len(Q_list) == 0:
                break

        if self.starting_node in self.community:
            return sorted(self.community)
        return []

    def compute_modularity(self, auxiliary_info, candidate_node):
        mode = auxiliary_info
        ind_s, outd_s = 0, 0

        community = list(self.community)
        if mode == "addition":
            community.append(candidate_node)
        elif mode == "deletion":
            community.remove(candidate_node)

        for node in community:
            for neighbor in self.graph.neighbors(node):
                if neighbor in community:
                    ind_s += 1
                else:
                    outd_s += 1

        return float(ind_s) / float(outd_s)

    def should_leave_boundary(self, possibly_leaving_node, neighbor_node):
        neighbors = set(self.graph.neighbors(possibly_leaving_node))
        neighbors.discard(neighbor_node)
        for neighbor in neighbors:
            if (neighbor in self.community) is False:
                return False
        return True



###############################################################################
#  Link Partitioning Around Medoids https://arxiv.org/abs/1907.08731          #
#  Alexander Ponomarenko, Leonidas Pitsoulis, Marat Shamshetdinov             #
#                                                                             #
# Contact us:                                                                 #
# aponomarenko@hse.ru - Alexander                                             #
# pitsouli@auth.gr    - Leonidas                                              #
###############################################################################

from collections import defaultdict
import networkx as nx
import numpy as np
from pyclustering.cluster.kmedoids import kmedoids
from cdlib import NodeClustering


def LPAM(graph, k=2, threshold=0.5, distance="amp", seed=0):
    """
    Link Partitioning Around Medoids

    :param graph: a networkx object
    :param k: number of clusters
    :param threshold: merging threshold in [0,1], default 0.5
    :param distance: type of distance: "amp" - amplified commute distance, or
    "cm" - commute distance, or distance matrix between all edges as np ndarray
    :param seed: random seed for k-medoid heuristic

    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.lpam(G, k=2, threshold=0.4, distance = "amp")

    :References:
    Link Partitioning Around Medoids https://arxiv.org/abs/1907.08731
    Alexander Ponomarenko, Leonidas Pitsoulis, Marat Shamshetdinov
    """

    def getCommuteDistace(G):
        """
        Returns commute distance matrix
        """
        verts = list(G.nodes)
        n = len(verts)
        vol = nx.volume(G, verts)

        # use NetworkX to get Laplacian
        L = nx.laplacian_matrix(G)
        L = L.todense()
        Gamma = L + (1 / n) * np.ones([n, n])
        CM = np.zeros([n, n])

        # get Moore-Penrose pseudo inverse
        Gamma_pinv = np.linalg.pinv(Gamma, rcond=1e-4)
        for i in range(n):
            for j in range(i + 1, n):
                CM[i, j] = vol * (
                    Gamma_pinv[i, i] + Gamma_pinv[j, j] - 2 * Gamma_pinv[i, j]
                )
                CM[j, i] = CM[i, j]
        return CM

    def getAmp(G):
        """
        Returns amplified commute distance matrix
        """
        verts = list(G.nodes)
        n = len(verts)

        # get adj matrix
        A = nx.adjacency_matrix(G)
        A = A.todense()

        # use NetworkX to get Laplacian
        L = nx.laplacian_matrix(G)
        L = L.todense()
        Gamma = L + (1 / n) * np.ones([n, n])
        C_AMP = np.zeros([n, n])

        # get Moore-Penrose pseudo inverse
        Gamma_pinv = np.linalg.pinv(Gamma, rcond=1e-4)
        for i in range(n):
            for j in range(i + 1, n):
                r_ij = (
                    Gamma_pinv[i, i] + Gamma_pinv[j, j] - 2 * Gamma_pinv[i, j]
                )  # resistance dist
                d_i = G.degree(list(G.nodes())[i])
                d_j = G.degree(list(G.nodes())[j])
                if d_i != 0 and d_j != 0:
                    s_ij = r_ij - (1 / d_i) - (1 / d_j)
                    w_ij = A[i, j]
                    w_ii = A[i, i]
                    w_jj = A[j, j]
                    u_ij = (
                        ((2 * w_ij) / (d_i * d_j))
                        - (w_ii / (d_i ** 2))
                        - (w_jj / (d_j ** 2))
                    )
                    C_AMP[i, j] = s_ij + u_ij
                    C_AMP[j, i] = s_ij + u_ij
                else:
                    C_AMP[i, j] = np.NaN
                    C_AMP[j, i] = np.NaN
        return C_AMP

    line_graph = nx.line_graph(graph)
    D = None
    distance_name = distance
    if distance == "amp":
        D = getAmp(line_graph)
    if distance == "cm":
        D = getCommuteDistace
    if isinstance(distance, np.ndarray):
        D = distance
        distance_name = "custom"
    if D is None:
        raise TypeError('Parameter distance should be "amp"/"cm", or numpy.ndarray')
    _n = len(line_graph.nodes())
    np.random.seed(0)
    initial_medoids = np.random.choice(_n, k, replace=False)
    kmedoids_instance = kmedoids(D, initial_medoids, data_type="distance_matrix")
    # run cluster analysis and obtain results
    kmedoids_instance.process()

    clusters = kmedoids_instance.get_clusters()

    final_clusters = {}
    for c_i, c in enumerate(clusters):
        for line_vertex in c:
            source, target = list(line_graph.nodes())[line_vertex]
            if source not in final_clusters:
                final_clusters[source] = []
            final_clusters[source].append(c_i)
            if target not in final_clusters:
                final_clusters[target] = []

            final_clusters[target].append(c_i)

    res_clusters = {}
    for v, l in final_clusters.items():
        degree = len(l)
        res = defaultdict(list)
        for x in l:
            res[x].append(x)
        covering = np.zeros(k)
        for c_i, _l in res.items():
            covering[c_i] = len(_l) / degree

        res_clusters[v] = covering

    _res_clusters = [[] for i in range(k)]

    for v, l in res_clusters.items():
        for i in range(k):
            if l[i] >= threshold:
                _res_clusters[i].append(v)

    return NodeClustering(
        communities=[c for c in _res_clusters if len(c) > 0],
        graph=graph,
        method_name="lpam " + distance_name,
        method_parameters={
            "k": k,
            "threshold": threshold,
            "distance": distance_name,
            "seed": seed,
        },
        overlap=True,
    )
